@@ -109,121 +109,157 @@
   (cond
    ((null? exc) "()")
    ((list? exc)
-    (cond
-     ((memq (car exc) '(duplicate-imported-declaration
-			duplicate-imported-mutable-declaration))
-      (let ((s-error-type (car exc)))
-	(case s-error-type
-	  ((duplicate-imported-declaration)
-	   (let ((s-name (cadr exc)))
-	     (string-append "Duplicate imported declaration of variable "
-			    (get-var-orig-name (symbol->string s-name)))))
-	  ((duplicate-imported-mutable-declaration)
-	   (let ((s-name (cadr exc)))
-	     (string-append
-	      "Duplicate imported declaration of mutable variable "
-	      (get-var-orig-name (symbol->string s-name)))))
-	  (else (string-append "??? ")))))
-     ((eq? (car exc) 'type-mismatch-in-param-proc-appl-1)
-      "Argument list type mismatch")
-     ((eq? (car exc) 'method-declaration-not-defined)
-      (string-append "A declaration of method "
-		     (symbol->string (cdr (assq 's-gen-proc (cdr exc))))
-		     " not defined"))
-     ((eq? (car exc) 'field-initializer-type-mismatch)
-      (string-append "Type mismatch in the initializer of field "
-		     (symbol->string (cdr (assq 's-field-name (cdr exc))))))
-     ((eq? (car exc) 'field-type-mismatch-in-param-class)
-      (string-append "Type mismatch in the initializer of field "
-		     (symbol->string (cdr (assq 's-field-name (cdr exc))))))
-     ((eq? (car exc) 'multiple-prim-class-definitions)
-      (string-append "Multiple primitive class definitions with membership "
-		     "predicate "
-		     (symbol->string (cadr exc))))
-     ((eq? (car exc) 'let:initializer-none)
-      (string-append
-       "Let initializer of type <none> for variable "
-       (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
-     ((eq? (car exc) 'letrec-initializer-with-type-none)
-      (string-append
-       "Letrec initializer of type <none> for variable "
-       (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
-     ((eq? (car exc) 'initializer-expression-with-type-none)
-      (string-append
-       "Initializer of type <none> for variable "
-       (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
-     ((eq? (car exc) 'let:type-mismatch)
-      (string-append
-       "Type mismatch for let variable "
-       (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
-     ((eq? (car exc) 'letrec-var-type-mismatch)
-      (string-append
-       "Type mismatch for letrec variable "
-       (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
-     ((eq? (car exc) 'invalid-let-var-type)
-      (string-append
-       "Invalid type for let variable "
-       (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
-     ((eq? (car exc) 'declaration-in-different-module)
-      (string-append
-       "Declaration of variable "
-       (symbol->string (get-var-orig-name (cdr (assq 's-var-name (cdr exc)))))
-       " in different module "
-       (convert-module-to-text (cdr (assq 'module (cdr exc))))))
-     ((eq? (car exc) 'illegal-use-of-keyword)
-      (string-append
-       "Illegal use of keyword "
-       (symbol->string (cdr (assq 's-keyword (cdr exc))))))
-     ((eq? (car exc) 'static-cast-type-mismatch)
-      "Type mismatch in static cast")
-     ((eq? (car exc) 'field-type-none1)
-      (string-append "The type of the field "
-		     (symbol->string (cdr (assq 's-field-name (cdr exc))))
-		     " is <none>"))
-     ((eq? (car exc) 'field-ref:nonexistent-field)
-      (string-append "Reference to nonexistent field "
-		     (symbol->string (cdr (assq 's-field-name (cdr exc))))))
-     ((eq? (car exc) 'field-set!:nonexistent-field)
-      (string-append "Trying to set nonexistent field "
-		     (symbol->string (cdr (assq 's-field-name (cdr exc))))))
-     ((eq? (car exc) 'nonexistent-field)
-      (string-append "Trying to access nonexistent field "
-		     (symbol->string (cdr (assq 's-field-name (cdr exc))))))
-     ((eq? (car exc) 'forward-def-type-mismatch)
-      (let ((s-name (cdr (assq 's-name (cdr exc)))))
+    (let ((x-head (car exc)))
+      (cond
+       ((memq x-head '(duplicate-imported-declaration
+			  duplicate-imported-mutable-declaration))
+	(let ((s-error-type x-head))
+	  (case s-error-type
+	    ((duplicate-imported-declaration)
+	     (let ((s-name (cadr exc)))
+	       (string-append "Duplicate imported declaration of variable "
+			      (get-var-orig-name (symbol->string s-name)))))
+	    ((duplicate-imported-mutable-declaration)
+	     (let ((s-name (cadr exc)))
+	       (string-append
+		"Duplicate imported declaration of mutable variable "
+		(get-var-orig-name (symbol->string s-name)))))
+	    (else (string-append "??? ")))))
+       ((eq? x-head 'type-mismatch-in-param-proc-appl-1)
+	"Argument list type mismatch")
+       ((eq? x-head 'method-declarations-not-defined)
+	(string-append "Declarations of the following methods not defined in"
+		       " body compilation:\n"
+		       (let* ((al-prop (cdr exc))
+			      (l-undefined (cdr (assq 'l-undefined al-prop))))
+			 (apply string-append
+				(map-in-order
+				 (lambda (p)
+				   (string-append
+				    (symbol->string (car p))
+				    ": "
+				    (target-object-as-string (cdr p))
+				    "\n"))
+				 l-undefined)))))
+       ((eq? x-head 'field-initializer-type-mismatch)
+	(string-append "Type mismatch in the initializer of field "
+		       (symbol->string (cdr (assq 's-field-name (cdr exc))))))
+       ((eq? x-head 'field-type-mismatch-in-param-class)
+	(string-append "Type mismatch in the initializer of field "
+		       (symbol->string (cdr (assq 's-field-name (cdr exc))))))
+       ((eq? x-head 'multiple-prim-class-definitions)
+	(string-append "Multiple primitive class definitions with membership "
+		       "predicate "
+		       (symbol->string (cadr exc))))
+       ((eq? x-head 'let:initializer-none)
 	(string-append
-	 "Type mismatch when defining declared variable "
-	 (symbol->string s-name))))
-     ((eq? (car exc) 'mutable-forward-def-type-mismatch)
-      (let ((s-name (cdr (assq 's-name (cdr exc)))))
+	 "Let initializer of type <none> for variable "
+	 (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
+       ((eq? x-head 'letrec-initializer-with-type-none)
 	(string-append
-	 "Type mismatch when defining declared variable "
-	 (symbol->string s-name))))
-     ((eq? (car exc) 'unaccessible-letrec-variable)
-      (let ((s-name (cdr (assq 's-name (cdr exc)))))
+	 "Letrec initializer of type <none> for variable "
+	 (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
+       ((eq? x-head 'initializer-expression-with-type-none)
 	(string-append
-	 "Unaccessible letrec variable "
-	 (symbol->string (get-var-orig-name s-name)))))
-     ((eq? (car exc) 'result-type-mismatch-1)
-      "Result type mismatch")
-     ((eq? (car exc) 'generic-static-dispatch-error-1)
-      "Generic static dispatch error")
-     ((memq (car exc)
-	    (list 'invalid-number-of-args 'invalid-number-of-args-2))
-      (let ((s-proc-name (cdr (assq 's-proc-name (cdr exc)))))
-	(if (not-null? s-proc-name)
+	 "Initializer of type <none> for variable "
+	 (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
+       ((eq? x-head 'let:type-mismatch)
+	(string-append
+	 "Type mismatch for let variable "
+	 (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
+       ((eq? x-head 'letrec-var-type-mismatch)
+	(string-append
+	 "Type mismatch for letrec variable "
+	 (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
+       ((eq? x-head 'invalid-let-var-type)
+	(string-append
+	 "Invalid type for let variable "
+	 (symbol->string (get-var-orig-name (cdr (assq 's-name (cdr exc)))))))
+       ((eq? x-head 'declaration-in-different-module)
+	(string-append
+	 "Declaration of variable "
+	 (symbol->string (get-var-orig-name (cdr (assq 's-var-name (cdr exc)))))
+	 " in different module "
+	 (convert-module-to-text (cdr (assq 'module (cdr exc))))))
+       ((eq? x-head 'illegal-use-of-keyword)
+	(string-append
+	 "Illegal use of keyword "
+	 (symbol->string (cdr (assq 's-keyword (cdr exc))))))
+       ((eq? x-head 'static-cast-type-mismatch)
+	"Type mismatch in static cast")
+       ((eq? x-head 'field-type-none1)
+	(string-append "The type of the field "
+		       (symbol->string (cdr (assq 's-field-name (cdr exc))))
+		       " is <none>"))
+       ((eq? x-head 'field-ref:nonexistent-field)
+	(string-append "Reference to nonexistent field "
+		       (symbol->string (cdr (assq 's-field-name (cdr exc))))))
+       ((eq? x-head 'field-set!:nonexistent-field)
+	(string-append "Trying to set nonexistent field "
+		       (symbol->string (cdr (assq 's-field-name (cdr exc))))))
+       ((eq? x-head 'nonexistent-field)
+	(string-append "Trying to access nonexistent field "
+		       (symbol->string (cdr (assq 's-field-name (cdr exc))))))
+       ((eq? x-head 'forward-def-type-mismatch)
+	(let ((s-name (cdr (assq 's-name (cdr exc)))))
+	  (string-append
+	   "Type mismatch when defining declared variable "
+	   (symbol->string s-name))))
+       ((eq? x-head 'mutable-forward-def-type-mismatch)
+	(let ((s-name (cdr (assq 's-name (cdr exc)))))
+	  (string-append
+	   "Type mismatch when defining declared variable "
+	   (symbol->string s-name))))
+       ((eq? x-head 'unaccessible-letrec-variable)
+	(let ((s-name (cdr (assq 's-name (cdr exc)))))
+	  (string-append
+	   "Unaccessible letrec variable "
+	   (symbol->string (get-var-orig-name s-name)))))
+       ((eq? x-head 'result-type-mismatch-1)
+	"Result type mismatch")
+       ((eq? x-head 'generic-static-dispatch-error-1)
+	"Generic static dispatch error")
+       ((memq x-head
+	      (list 'invalid-number-of-args 'invalid-number-of-args-2))
+	(let ((s-proc-name (cdr (assq 's-proc-name (cdr exc)))))
+	  (if (not-null? s-proc-name)
+	      (string-append
+	       "Invalid number of arguments for procedure "
+	       (symbol->string s-proc-name))
+	      "Invalid number of arguments")))
+       ((eq? x-head 'tuple-ref:not-a-tuple)
+	(let ((tt (cdr (assq 'tt (cdr exc)))))
+	  (string-append
+	   "Type " (target-object-as-string tt) " is not a tuple")))
+       ((eq? x-head 'declarations-not-defined)
+	(let ((l-undefined (cdr (assq 'l-undefined (cdr exc)))))
+	  (cond
+	   ((null? l-undefined) "Internal error")
+	   ((null? (cdr l-undefined))
+	    (string-append "Declaration of variable "
+			   (if (symbol? (car l-undefined))
+			       (symbol->string (car l-undefined))
+			       "?")
+			   " not defined in target body compilation"))
+	   (else
 	    (string-append
-	     "Invalid number of arguments for procedure "
-	     (symbol->string s-proc-name))
-	    "Invalid number of arguments")))
-     (else
-      (if (= (length exc) 1)
-	  (if (symbol? (car exc))
-	      (string-append "Error " (symbol->string (car exc)))
-	      "?")
-	  (if (symbol? (car exc))
-	      (string-append "Error (" (symbol->string (car exc)) " ...)")
-	      "?")))))
+	     "Declarations of the following variables are undefined"
+	     " in the body compilation:\n"
+	     (apply string-append
+		    (map-in-order
+		     (lambda (x)
+		       (if (symbol? x)
+			   (string-append "  " (symbol->string x) "\n")
+			   "  ?\n"))
+		     l-undefined)))))))
+       (else
+	(if (= (length exc) 1)
+	    (if (symbol? x-head)
+		(string-append "Error " (symbol->string x-head))
+		"?")
+	    (if (symbol? x-head)
+		(string-append "Error (" (symbol->string x-head) " ...)")
+		"?"))))))
    (else
     (cond
      ((eq? exc 'module-name-mismatch)
@@ -235,14 +271,6 @@
 		     (get-var-orig-name
 		      (symbol->string (hfield-ref compiler 'error-info)))))
      ;; Global variables don't need get-var-orig-name.
-     ((eqv? exc 'declaration-not-defined)
-      (string-append "Declaration of variable "
-		     (symbol->string (hfield-ref compiler 'error-info))
-		     " not defined"))
-     ((eqv? exc 'method-declaration-not-defined)
-      (string-append "A declaration of method "
-		     (symbol->string (hfield-ref compiler 'error-info))
-		     " not defined"))
      ((eqv? exc 'duplicate-definition)
       (string-append "Duplicate definition"
 		     (get-var-text (hfield-ref compiler 'error-info))))
@@ -480,10 +508,10 @@
 			   " from module "
 			   module-text
 			   "."))
-	    (string-append error-text
-			   " while importing module "
-			   module-text
-			   ".")))))
+	  (string-append error-text
+			 " while importing module "
+			 module-text
+			 ".")))))
 
 
 (define (make-interface-error-message compiler exc)
@@ -504,44 +532,48 @@
   (let ((error-text (get-error-text compiler exc))
 	(repr (hfield-ref compiler 'current-repr))
 	(toplevel-repr (hfield-ref compiler 'current-toplevel-repr)))
-    (if (eqv? repr toplevel-repr)
-	(if (not-null? toplevel-repr)
-	    (string-append error-text
-			   " while compiling "
-			   (get-repr-text
-			    toplevel-repr
-			    #t)
-			   " for the target body.")
-	    (string-append error-text
-			   " in target body compilation."))
-	(cond
-	 ((and (not-null? repr) (not-null? toplevel-repr))
-	  (string-append error-text
-			 " while compiling "
-			 (get-repr-text
-			  repr
-			  #f)
-			 " in "
-			 (get-repr-text
-			  toplevel-repr
-			  #t)
-			 " for the target body."))
-	 ((and (null? repr) (not-null? toplevel-repr))
+    (cond
+     ((and (pair? exc) (memq (car exc)
+			     (list 'declarations-not-defined
+				   'method-declarations-not-defined)))
+      error-text)
+     ((eqv? repr toplevel-repr)
+      (if (not-null? toplevel-repr)
 	  (string-append error-text
 			 " while compiling "
 			 (get-repr-text
 			  toplevel-repr
 			  #t)
-			 " for the target body."))
-	 ((and (not-null? repr) (null? toplevel-repr))
+			 " for the target body.")
 	  (string-append error-text
-			 " while compiling "
-			 (get-repr-text
-			  repr
-			  #f)
-			 " for the target body."))
-	 (else (string-append error-text
-			      " while compiling the target body."))))))
+			 " in target body compilation.")))
+     ((and (not-null? repr) (not-null? toplevel-repr))
+      (string-append error-text
+		     " while compiling "
+		     (get-repr-text
+		      repr
+		      #f)
+		     " in "
+		     (get-repr-text
+		      toplevel-repr
+		      #t)
+		     " for the target body."))
+     ((and (null? repr) (not-null? toplevel-repr))
+      (string-append error-text
+		     " while compiling "
+		     (get-repr-text
+		      toplevel-repr
+		      #t)
+		     " for the target body."))
+     ((and (not-null? repr) (null? toplevel-repr))
+      (string-append error-text
+		     " while compiling "
+		     (get-repr-text
+		      repr
+		      #f)
+		     " for the target body."))
+     (else (string-append error-text
+			  " while compiling the target body.")))))
 
 
 (define (make-other-error-message compiler exc)
@@ -551,11 +583,11 @@
 
 (define (get-list-error-message exc)
   (let ((s-error-type (car exc)))
-  (cond
-   ((memq s-error-type '(noncovariant-method-definition
-			 noncovariant-method-declaration))
-    (get-noncov-method-error-text exc))
-   (else "Error."))))
+    (cond
+     ((memq s-error-type '(noncovariant-method-definition
+			   noncovariant-method-declaration))
+      (get-noncov-method-error-text exc))
+     (else "Error."))))
 
 
 (define (delete-target-file compiler)
@@ -615,7 +647,7 @@
       (display message)
       (newline)
       (if (and 
-;;	   (eqv? (hfield-ref compiler 'state) 'parsing)
+	   ;;	   (eqv? (hfield-ref compiler 'state) 'parsing)
 	   (hfield-ref compiler 'display-erroneous-expr?)
 	   (not-null? (hfield-ref compiler 'current-expr)))
 	  (begin
@@ -662,12 +694,12 @@
 	      (display "Needed type variables: ")
 	      (display (target-object-as-string l-needed))
 	      (newline)))
-	   ((eq? (car exc) 'method-declaration-not-defined)
-	    (let* ((al-prop (cdr exc))
-		   (tt-method (cdr (assq 'tt-method al-prop))))
-	      (display "Declared type: ")
-	      (display (target-object-as-string tt-method))
-	      (newline)))
+	   ;; ((eq? (car exc) 'method-declaration-not-defined)
+	   ;;  (let* ((al-prop (cdr exc))
+	   ;; 	   (tt-method (cdr (assq 'tt-method al-prop))))
+	   ;;    (display "Declared type: ")
+	   ;;    (display (target-object-as-string tt-method))
+	   ;;    (newline)))
 	   ((memq (car exc)
 		  (list 'redecl-type-mismatch
 			'forward-def-type-mismatch
