@@ -14,6 +14,7 @@
 (define rebind-variable-fwd! '())
 (define make-tpci-pair-fwd '())
 (define theme-class-of-fwd '())
+(define contains-free-tvars-fwd? '())
 
 
 (define-hrecord-type <entity> ()
@@ -428,6 +429,33 @@
    #f))
 
 
+(define (make-normal-variable9 address type type-dispatched? exact-type?
+			       read-only? value)
+  (make-normal-variable0 address type type-dispatched? exact-type? read-only?
+			 #f #f #f value '() #f #f))
+
+
+(define (make-normal-variable10 address type type-dispatched?
+				exact-type? read-only?
+				volatile?
+				forward-decl?
+				value exported?)
+  (dwl4 "make-normal-variable10")
+  (make-normal-variable0
+   address
+   type
+   type-dispatched?
+   exact-type?
+   read-only?
+   volatile?
+   forward-decl?
+   #f
+   value
+   '()
+   exported?
+   #f))
+
+
 (define (get-entity-value entity)
   (cond
    ((is-target-object? entity) entity)
@@ -464,6 +492,12 @@
   (hfield-ref ent 'type-dispatched?))
 
 
+(define (entity-type-dispatched2? ent)
+  (assert (hrecord-is-instance? ent <entity>))
+  (and (hfield-ref ent 'type-dispatched?)
+       (not (contains-free-tvars-fwd? ent))))
+
+
 (define (is-forward-decl-entity? ent)
   (or
    (and (is-target-object? ent) (hfield-ref ent 'incomplete?))
