@@ -157,12 +157,13 @@
 
 ;; The following procedure should also work for general procedure
 ;; applications.
+;; Argument runtime-arglist-typecheck? is not used.
 (define (translate-simple-proc-appl-expression binder
 					       proc arg-list static-arg-types
 					       type-check?
 					       inside-param-def?
 					       runtime-arglist-typecheck?)
-  (dwl3 "translate-simple-proc-appl-expression")
+  (dwl2 "translate-simple-proc-appl-expression")
   (assert (is-binder? binder))
   (assert (is-entity? proc))
   (assert (list? arg-list))
@@ -173,42 +174,21 @@
   (assert (boolean? type-check?))
   (assert (boolean? inside-param-def?))
   (assert (boolean? runtime-arglist-typecheck?))
-  (dwl4 "translate-simple-proc-appl-expression/0")
+  (dwl2 "translate-simple-proc-appl-expression/0")
   (dvar1-set! proc)
-
-  ;; TBR
-  ;; (set! gl-counter15 (+ gl-counter15 1))
-  ;; (dwl3 gl-counter15)
-  ;; (if (is-target-object? proc)
-  ;;     (dwl3 (hfield-ref (hfield-ref proc 'address) 'source-name)))
-  ;; (if (= gl-counter15 79)
-  ;;     (begin
-  ;; 	(dvar1-set! binder)
-  ;; 	(dvar2-set! proc)
-  ;; 	(dvar3-set! arg-list)
-  ;; 	(dwl4 type-check?)
-  ;; 	(dwl4 inside-param-def?)
-  ;; 	(raise 'stop15)))
-  (if (and
-       (not-null? (hfield-ref proc 'address))
-       (eqv? (hfield-ref (hfield-ref proc 'address) 'source-name)
-	     'real->string))
-      (begin
-  	(dwl2 "real->string HEP")
-  	(dwl2 type-check?)
-	(dwl2 runtime-arglist-typecheck?)
-  	(dwl2 inside-param-def?)))
-
   (let* ((proc-type (get-entity-type proc))
-	 (tmp1 (begin (dwl4 "translate-simple-proc-appl-expression/0-1") 0))
+	 (tmp1 (begin (dwl2 "translate-simple-proc-appl-expression/0-1") 0))
 	 (static-arg-types1
 	  (if (eq? static-arg-types #f)
 	      (map get-expr-type arg-list)
 	      static-arg-types))
-	 (type-dispatched?
-	  (and (entity-type-dispatched? proc)
-	       (and-map? entity-type-dispatched? arg-list)))
-	 (tmp2 (begin (dwl4 "translate-simple-proc-appl-expression/0-2")
+	 ;; (type-dispatched?
+	 ;;  (and (entity-type-dispatched? proc)
+	 ;;       (and-map? entity-type-dispatched? arg-list)))
+	 (type-dispatched? (entity-type-dispatched? proc))
+	 (arg-list-type-dispatched?
+	  (and-map? entity-type-dispatched? arg-list))
+	 (tmp2 (begin (dwl2 "translate-simple-proc-appl-expression/0-2")
 		      (dvar1-set! proc)
 		      0))
 	 (always-returns?
@@ -216,7 +196,7 @@
 	   (tno-field-ref proc-type 'appl-always-returns?)
 	   (entity-always-returns? proc)
 	   (and-map? entity-always-returns? arg-list)))
-	 (tmp3 (begin (dwl4 "translate-simple-proc-appl-expression/0-3") 0))
+	 (tmp3 (begin (dwl2 "translate-simple-proc-appl-expression/0-3") 0))
 	 (never-returns?
 	  (or
 	   (tno-field-ref proc-type 'appl-never-returns?)
@@ -226,29 +206,8 @@
 	 (pure-args?
 	  (and-map? is-pure-entity? arg-list))
 	 (pure? (and pure-proc? pure-args?)))
-    (dwl4 "translate-simple-proc-appl-expression/0-1.5")
+    (dwl2 "translate-simple-proc-appl-expression/0-1.5")
     (dvar1-set! proc)
-
-    ;; TO BE REMOVED
-    ;; (if (and (hrecord-is-instance? proc <variable-reference>)
-    ;; 	 (var-ref-to-sym? proc 'myproc))
-    ;;     (begin
-    ;;       (dwl4 type-check?)
-    ;;       (dwl4 inside-param-def?)
-    ;;       (dvar1-set! proc)
-    ;;       (dvar2-set! arg-list)
-    ;;       (dvar3-set! binder)
-    ;;       (dvar4-set! v-proc-type)
-    ;;       (raise 'stop-var)))
-
-    ;; TBR
-    ;; (set! gl-counter2 (+ gl-counter2 1))
-    ;; (dwl4 gl-counter2)
-    ;; (if (= gl-counter2 13)
-    ;;     (begin
-    ;;       (dvar1-set! arg-list-type)
-    ;;       (dvar2-set! result-type)
-    ;;       (dvar3 arg-list)))
     (let* ((arg-list-type (tno-field-ref proc-type 'type-arglist))
 	   (l-arg-types (map get-entity-type arg-list))
 	   (l-sgn (pick-signatures l-arg-types))
@@ -282,17 +241,19 @@
       ;;  	(dvar2-set! result-type)
       ;; 	(dvar3-set! tt-actual-arglist)
       ;; 	(dvar4-set! binder)
-      ;; 	(dwl4 type-check?)
+      ;; 	(dwl2 type-check?)
       ;;      	(raise 'stop1)))
 
-      (dwl4 "translate-simple-proc-appl-expression/1")
+      (dwl2 "translate-simple-proc-appl-expression/1")
       (cond
-       ((and type-check? (not type-dispatched?) (not inside-param-def?))
+       ((and type-check? (not inside-param-def?)
+	     (or (not type-dispatched?)
+		 (not arg-list-type-dispatched?)))
 	(dvar1-set! proc)
 	(dvar2-set! arg-list)
 	(raise 'illegal-nondispatched-proc-appl))
        ((not type-dispatched?)
-	(dwl4 "translate-simple-proc-appl-expression/2")
+	(dwl2 "translate-simple-proc-appl-expression/2")
 	(make-hrecord <proc-appl>
 		      result-type
 		      #f
@@ -307,14 +268,14 @@
 		      proc
 		      arg-list
 		      '()
-		      static-arg-types1
+		      '()
 		      #t
 		      '()))
        ((and contains-sgn? arg-free-tvars?)
-	(dwl4 "translate-simple-proc-appl-expression/2")
+	(dwl2 "translate-simple-proc-appl-expression/2")
 	(make-hrecord <proc-appl>
 		      result-type
-		      #f
+		      #t
 		      #f
 		      '()
 		      pure?
@@ -331,24 +292,21 @@
 		      '()))
        ((or (not type-check?)
 	    correct-sgn?
+	    (not arg-list-type-dispatched?)
 	    (and inside-param-def?
-		 (or
-		  (contains-free-tvars-fwd? tt-actual-arglist)
-		  (contains-free-tvars-fwd? arg-list-type)))
-	    (begin (dwl4 "translate-simple-proc-appl-expression/1-2")
-		   #f)
-	    (check-arglist-types? binder
-				  tt-actual-arglist arg-list-type))
-	(dwl4 "translate-simple-proc-appl-expression/2")
+	    	 (or
+	    	  (contains-free-tvars-fwd? tt-actual-arglist)
+	    	  (contains-free-tvars-fwd? arg-list-type))))
+	;; Maybe we could omit the runtime typecheck in case correct-sgn?
+	;; is #t.
 	(let ((final?
 	       (and (is-t-instance? binder result-type
 				    tc-class)
 		    (not
 		     (tno-field-ref result-type 'inheritable?)))))
-	  (dwl4 "translate-simple-proc-appl-expression EXIT 1")
 	  (make-hrecord <proc-appl>
 			result-type
-			type-dispatched?
+			#t
 			final?
 			'()
 			pure?
@@ -362,20 +320,36 @@
 			'()
 			static-arg-types1
 			runtime-arglist-typecheck?
+			'())))	
+       ((check-arglist-types? binder
+			      tt-actual-arglist arg-list-type)
+	(let ((final?
+	       (and (is-t-instance? binder result-type
+				    tc-class)
+		    (not
+		     (tno-field-ref result-type 'inheritable?)))))
+	  (make-hrecord <proc-appl>
+			result-type
+			#t
+			final?
+			'()
+			pure?
+			#f
+			(not type-check?)
+			'()
+			always-returns?
+			never-returns?
+			proc
+			arg-list
+			'()
+			static-arg-types1
+			#f
 			'())))
        (else
 	(dvar1-set! binder)
 	(dvar2-set! tt-actual-arglist)
 	(dvar3-set! arg-list-type)
 	(dvar4-set! (list proc arg-list))
-
-	;; TBR
-	(dwl2 "HEP9")
-	(set! gl-counter9 (+ gl-counter9 1))
-	(dwl2 gl-counter9)
-	(dvar1-set! proc)
-	;;	    (raise 'stop9)
-
 	(raise
 	 (let* ((to-proc (get-entity-value proc))
 		(s-proc-name
@@ -397,15 +371,6 @@
 	   (list 'type-mismatch-in-proc-appl s-proc-name
 		 (cons 'actual-type tt-actual-arglist)
 		 (cons 'declared-type arg-list-type)))))))))
-;; (if (is-tc-param-proc? (get-entity-type proc))
-;; 	 (list 'type-mismatch-in-proc-appl
-;; 	       (tno-field-ref proc 'name))
-;; 	 'type-mismatch-in-proc-appl))))))))
-;; (let ((address (hfield-ref proc 'address)))
-;;   (if (not-null? address)
-;; 	   (list 'type-mismatch-in-proc-appl
-;; 		 (hfield-ref address 'source-name))
-;; 	   'type-mismatch-in-proc-appl)))))))))
 
 
 (define (translate-simple-proc-appl-expression1 binder
@@ -522,64 +487,48 @@
   (let ((result-type (tno-field-ref r-inst-type 'type-result))
 	(arg-list-type (tno-field-ref r-inst-type 'type-arglist))
 	(tt-actual-arglist (get-arglist-type-from-list r-arg-types)))
-
-    ;; TBR
-    ;; (if gl-flag13?
-    ;; 	(begin
-    ;; 	  (dvar1-set! binder)
-    ;; 	  (dvar2-set! tt-actual-arglist)
-    ;; 	  (dvar3-set! arg-list-type)
-    ;; 	  (raise 'stop13)))
-  ;; (set! gl-counter24 (+ gl-counter24 1))
-  ;; (dwl2 gl-counter24)
-  ;; (if (= gl-counter24 5)
-  ;;     (begin
-  ;; 	(dvar1-set! binder)
-  ;; 	(dvar2-set! tt-actual-arglist)
-  ;; 	(dvar3-set! arg-list-type)
-  ;; 	(raise 'stop24)))
-
-  (cond
-   ((null? tt-actual-arglist)
-    (raise 'internal-null-argument-list))
-   ((check-arglist-types?
-     binder tt-actual-arglist arg-list-type)
-    (let ((static-arg-types (map get-expr-type arglist))
-	  (exact-type?
-	   (and
-	    (not (is-t-type-variable? result-type))
-	    (is-t-instance? binder result-type tc-class)
-	    (not (tno-field-ref result-type 'inheritable?)))))
-      ;; need-revision? = #f
-      (make-hrecord <proc-appl>
-		    result-type
-		    #t exact-type? '()
-		    pure? #f #f '()
-		    always-returns? never-returns?
-		    e-proc arglist r-type-var-values
-		    static-arg-types
-		    #f
-		    l-default-params)))
-   (else
-    (dvar1-set! tt-actual-arglist)
-    (dvar2-set! arg-list-type)
-    (dvar3-set! e-proc)
-    (dvar4-set! arglist)
-    (raise (list 'type-mismatch-in-param-proc-appl-1
-		 (cons 'actual-type tt-actual-arglist)
-		 (cons 'declared-type arg-list-type)))))))
+    (cond
+     ((null? tt-actual-arglist)
+      (raise 'internal-null-argument-list))
+     ((check-arglist-types?
+       binder tt-actual-arglist arg-list-type)
+      ;;    (let ((static-arg-types (map get-expr-type arglist))
+      (let ((exact-type?
+	     (and
+	      (not (is-t-type-variable? result-type))
+	      (is-t-instance? binder result-type tc-class)
+	      (not (tno-field-ref result-type 'inheritable?)))))
+	;; need-revision? = #f
+	(make-hrecord <proc-appl>
+		      result-type
+		      #t exact-type? '()
+		      pure? #f #f '()
+		      always-returns? never-returns?
+		      e-proc arglist r-type-var-values
+		      ;; static-arg-types
+		      r-arg-types
+		      #f
+		      l-default-params)))
+     (else
+      (dvar1-set! tt-actual-arglist)
+      (dvar2-set! arg-list-type)
+      (dvar3-set! e-proc)
+      (dvar4-set! arglist)
+      (raise (list 'type-mismatch-in-param-proc-appl-1
+		   (cons 'actual-type tt-actual-arglist)
+		   (cons 'declared-type arg-list-type)))))))
 
 
 (define (translate-param-proc-appl-inside-param binder
 						e-proc arglist
+						type-dispatched?
 						r-inst-type r-arg-types
 						r-type-var-values
 						pure?
 						always-returns?
 						never-returns?
 						l-default-params)
-  ;; We know that the application type has been dispatched.
-  (dwl2 "translate-param-proc-appl-inside-param")
+  (dwl1 "translate-param-proc-appl-inside-param")
   (assert (is-binder? binder))
   (assert (is-entity? r-inst-type))
   (assert (and (list? r-type-var-values)
@@ -589,59 +538,44 @@
   (assert (boolean? never-returns?))
   (assert (and (list? l-default-params)
 	       (and-map? is-entity? l-default-params)))
-  (dwl2 "translate-param-proc-appl-inside-param/1")
-  (if (let* ((result-type (tno-field-ref r-inst-type 'type-result))
-	     (arg-list-type (tno-field-ref r-inst-type 'type-arglist))
-	     (tt-actual-arglist (get-arglist-type-from-list r-arg-types)))
-	(dwl2 "translate-param-proc-appl-inside-param/2")
-
-	;; TBR
-	(set! gl-counter17 (+ gl-counter17 1))
-	(dwl3 gl-counter17)
-	(if (= gl-counter17 3)
-	    (begin
-	      (dvar1-set! tt-actual-arglist)
-	      (dvar2-set! arg-list-type)
-	      (dvar3-set! e-proc)
-	      (dvar4-set! arglist)))
-;;	      (raise 'stop17)))
-
-	(cond
-	 ((null? tt-actual-arglist) #f)
-	 (else (check-arglist-types?
-		binder tt-actual-arglist arg-list-type))))
-      (let* ((result-type (tno-field-ref r-inst-type 'type-result))
-	     (exact-type? (is-final-class? binder result-type)))
-	(dwl2 "translate-param-proc-appl-inside-param/3")
-	(make-hrecord
-	 <proc-appl>
-	 result-type
-	 #t exact-type? '()
-	 pure? #f #t
-	 '() always-returns? never-returns?
-	 e-proc arglist '()
-	 (map get-entity-type arglist)
-	 #f
-	 l-default-params))
-      (if (or (contains-type-variables-fwd?
-	       (tno-field-ref r-inst-type 'type-arglist))
-	      (contains-type-variables-fwd? r-arg-types))
-	  (let* ((result-type (tno-field-ref r-inst-type 'type-result))
-		 (exact-type? (is-final-class? binder result-type)))
-	    (dwl2 "translate-param-proc-appl-inside-param/4")
-	    (make-hrecord
-	     <proc-appl>
-	     result-type #t exact-type? '()
-	     pure? #f #t
-	     '() always-returns? never-returns?
-	     e-proc arglist '()
-	     (map get-entity-type arglist)
-	     #f
-	     l-default-params))
-	  (begin
-	    (dvar1-set! r-inst-type)
-	    (dvar2-set! r-arg-types)
-	    (raise 'param-proc-appl-type-mismatch)))))
+  (let* ((arg-list-type (tno-field-ref r-inst-type 'type-arglist))
+	 (tt-actual-arglist (get-arglist-type-from-list r-arg-types))
+	 (result-type (tno-field-ref r-inst-type 'type-result))
+	 (exact-type? (is-final-class? binder result-type))
+	 (free-tvars?
+	  (or
+	   (contains-type-variables-fwd? arg-list-type)
+	   (contains-type-variables-fwd? tt-actual-arglist))))
+    (cond
+     ((and
+       type-dispatched?
+       (not free-tvars?)
+       (not (check-arglist-types?
+	     binder tt-actual-arglist arg-list-type)))
+      (raise 'param-proc-appl-type-mismatch))
+     (type-dispatched?
+      (make-hrecord
+       <proc-appl>
+       result-type
+       type-dispatched? exact-type? '()
+       pure? #f #t
+       '() always-returns? never-returns?
+       e-proc arglist '()
+       ;; (map get-entity-type arglist)
+       r-arg-types
+       #f
+       l-default-params))
+     (else
+      (make-hrecord
+       <proc-appl>
+       result-type
+       type-dispatched? exact-type? '()
+       pure? #f #t
+       '() always-returns? never-returns?
+       e-proc arglist '()
+       (map get-entity-type arglist)
+       #t
+       l-default-params)))))
 
 
 (define (make-general-gen-proc-appl binder gen-proc
@@ -651,7 +585,7 @@
 				    always-returns?
 				    never-returns?
 				    type-check?)
-  (dwl4 "make-general-gen-proc-appl")
+  (dwl1 "make-general-gen-proc-appl")
   (assert (is-binder? binder))
   (assert (is-entity? gen-proc))
   (assert (and (list? arglist)
@@ -765,12 +699,161 @@
 	'())))
 
 
+(define (do-translate-param-proc-appl0 binder inside-param-def?
+				       type-dispatched?
+				       e-proc arglist l-default-params
+				       ppc inst-type
+				       pure? always-returns? never-returns?)
+  (assert (not-null? inst-type))
+  (let* ((arg-types (map get-entity-type arglist))
+	 (type-vars (tno-field-ref ppc 'l-tvars))
+	 ;; We make sure that the type variables in the instance
+	 ;; type are unique, i.e. they don't appear in the
+	 ;; procedure expression or in the arguments.
+	 ;; Does calling contains-specified-tvars-fwd? have
+	 ;; any reason?
+	 (uniq
+	  (if (contains-specified-tvars-fwd? inst-type
+					     type-vars)
+	      (make-tvars-unique binder type-vars inst-type)
+	      (cons inst-type type-vars)))
+	 (inst-type1 (car uniq))
+	 (new-type-vars (cdr uniq))
+	 (new-arg-types (make-tvars-unique2 binder arg-types))
+	 (src-tvars (get-all-tvars new-arg-types))
+	 (arg-list-type0
+	  (tno-field-ref inst-type1 'type-arglist))
+	 (all-tvars0 (append src-tvars new-type-vars))
+	 (fixed-tvars (hfield-ref binder 'fixed-tvars))
+	 (all-tvars
+	  (get-nonfixed-tvars all-tvars0 fixed-tvars)))
+    (let ((deductions (get-new-type-var-assoc-table)))
+      (deduce-argument-types binder deductions
+			     all-tvars
+			     new-arg-types arg-list-type0)
+      (if (not inside-param-def?)
+	  (cond
+	   ((all-tvars-correct? deductions new-type-vars)
+	    (let* ((tvar-bindings
+		    (hfield-ref deductions 'bindings))
+		   (r-inst-type
+		    (bind-type-vars binder tvar-bindings
+				    inst-type1))
+		   (r-type-var-values
+		    (get-type-var-values-from-deductions2
+		     new-type-vars deductions))
+		   (r-arg-types
+		    (bind-type-vars binder tvar-bindings
+				    new-arg-types)))
+	      (translate-normal-param-proc-appl2
+	       binder e-proc arglist r-inst-type
+	       r-arg-types
+	       r-type-var-values pure? always-returns?
+	       never-returns?
+	       l-default-params)))
+	   ((not-null? l-default-params)
+	    (let* ((tvar-bindings
+		    (map cons new-type-vars l-default-params))
+		   (r-inst-type
+		    (bind-type-vars binder tvar-bindings
+				    inst-type1))
+		   (r-arg-types
+		    (bind-type-vars binder tvar-bindings
+				    new-arg-types)))
+	      (translate-normal-param-proc-appl2
+	       binder e-proc arglist r-inst-type
+	       r-arg-types
+	       l-default-params pure? always-returns?
+	       never-returns?
+	       l-default-params)))
+	   (else
+	    (let ((s-name
+		   (if (not-null? (hfield-ref e-proc 'address))
+		       (hfield-ref (hfield-ref e-proc 'address)
+				   'source-name)
+		       '())))
+	      (raise (list 'did-not-deduce-all-type-vars
+			   s-name
+			   (cons 'actual-type new-arg-types)
+			   (cons 'declared-type arg-list-type0)
+			   (cons
+			    'bindings
+			    (hfield-ref deductions 'bindings))
+			   (cons 'needed new-type-vars))))))
+	  (cond
+	   ((all-tvars-correct? deductions new-type-vars)
+	    (let* ((tvar-bindings
+		    (hfield-ref deductions 'bindings))
+		   (r-inst-type
+		    (bind-type-vars binder tvar-bindings
+				    inst-type1))
+		   (r-type-var-values
+		    (get-type-var-values-from-deductions2
+		     new-type-vars deductions))
+		   (r-arg-types
+		    (bind-type-vars binder tvar-bindings
+				    new-arg-types)))
+	      (translate-param-proc-appl-inside-param
+	       binder e-proc arglist type-dispatched?
+	       r-inst-type
+	       r-arg-types
+	       r-type-var-values pure?
+	       always-returns? never-returns?
+	       l-default-params)))
+	   ((not-null? l-default-params)
+	    (let* ((tvar-bindings
+		    (map cons new-type-vars l-default-params))
+		   (r-inst-type
+		    (bind-type-vars binder tvar-bindings
+				    inst-type1))
+		   (r-arg-types
+		    (bind-type-vars binder tvar-bindings
+				    new-arg-types)))
+	      (translate-param-proc-appl-inside-param
+	       binder e-proc arglist type-dispatched?
+	       r-inst-type
+	       r-arg-types
+	       l-default-params pure?
+	       always-returns? never-returns?
+	       l-default-params)))
+	   (else
+	    (let* ((tvar-bindings
+		    (hfield-ref deductions 'bindings))
+		   (r-inst-type
+		    (bind-type-vars binder tvar-bindings
+				    inst-type1))
+		   (result-type
+		    (tno-field-ref r-inst-type
+				   'type-result))
+		   (static-arg-types (map get-expr-type arglist)))
+	      (if (contains-only-specified-unbound-tvars-fwd?
+		   result-type
+		   fixed-tvars)
+		  (begin
+		    (make-hrecord
+		     <proc-appl>
+		     result-type
+		     #t #f '()
+		     pure? #f #t
+		     '() always-returns? never-returns?
+		     e-proc arglist '() static-arg-types
+		     #t
+		     l-default-params))
+		  (begin
+		    (make-hrecord
+		     <proc-appl>
+		     tc-object
+		     #f #f '()
+		     pure? #f #t
+		     '() always-returns? never-returns?
+		     e-proc arglist '() static-arg-types
+		     #t
+		     l-default-params))))))))))
+
+
 (define (do-translate-param-proc-appl1 binder inside-param-def?
 				       e-proc arglist l-default-params)
   (dwl2 "do-translate-param-proc-appl1")
-  (let ((address (hfield-ref e-proc 'address)))
-    (if (not-null? address)
-	(dwl2 (hfield-ref address 'source-name))))
   (assert (is-binder? binder))
   (assert (boolean? inside-param-def?))
   (assert (is-entity? e-proc))
@@ -784,8 +867,7 @@
 	      'procedure-argument-with-type-none
 	      (cons 's-proc-name
 		    (get-proc-name2 e-proc))))
-      (let ((t-proc (get-entity-value e-proc))
-	    (static-arg-types (map get-expr-type arglist)))
+      (let ((t-proc (get-entity-value e-proc)))
 	(dwl2 "do-translate-param-proc-appl1/1")
 	;; Maybe we should remove the following test.
 	(if (null? t-proc)
@@ -814,6 +896,7 @@
 	      ;; (display "arg list: ")
 	      ;; (display (target-object-as-string arg-list-type00))
 	      ;; (newline)
+	      (dwl2 "do-translate-param-proc-appl1/2")
 	      (cond
 	       ((is-tuple-type? binder arg-list-type00)
 		(if (not (= (length arglist)
@@ -850,174 +933,26 @@
 		      (raise (list 'invalid-number-of-args-2
 				   (cons 's-proc-name
 					 (get-proc-name2 e-proc))))))))
-	      (dwl2 "do-translate-param-proc-appl1/2")
-	      (assert (not-null? inst-type))
-	      (dwl2 "do-translate-param-proc-appl1/3")
-	      (let* ((arg-types (map get-entity-type arglist))
-		     (type-vars (tno-field-ref ppc 'l-tvars))
-		     ;; We make sure that the type variables in the instance
-		     ;; type are unique, i.e. they don't appear in the
-		     ;; procedure expression or in the arguments.
-		     ;; Does calling contains-specified-tvars-fwd? have
-		     ;; any reason?
-		     (uniq
-		      (if (contains-specified-tvars-fwd? inst-type
-							 type-vars)
-			  (make-tvars-unique binder type-vars inst-type)
-			  (cons inst-type type-vars)))
-		     (tmp1 (begin (dwl2 "do-translate-param-proc-appl1/4") 0))
-		     (inst-type1 (car uniq))
-		     (new-type-vars (cdr uniq))
-		     (new-arg-types (make-tvars-unique2 binder arg-types))
-		     (tmp1 (begin (dwl2 "do-translate-param-proc-appl1/5") 0))
-		     (src-tvars (get-all-tvars new-arg-types))
-		     (arg-list-type0
-		      (tno-field-ref inst-type1 'type-arglist))
-		     (all-tvars0 (append src-tvars new-type-vars))
-		     (fixed-tvars (hfield-ref binder 'fixed-tvars))
-		     (all-tvars
-		      (get-nonfixed-tvars all-tvars0 fixed-tvars)))
-		(dwl2 "do-translate-param-proc-appl1/6")
-		(let ((deductions (get-new-type-var-assoc-table)))
-		  (dwl2 "do-translate-param-proc-appl1/7")
-		  (deduce-argument-types binder deductions
-		  			 all-tvars
-		  			 new-arg-types arg-list-type0)
-		  (dwl2 "do-translate-param-proc-appl1/8")
-		  (dwl2 "do-translate-param-proc-appl1/9")
+	      (if (not type-dispatched?)
 		  (if (not inside-param-def?)
-		      (cond
-		       ((not type-dispatched?)
-			(raise 'illegal-nondispatched-type))
-		       ((all-tvars-correct? deductions new-type-vars)
-			(dwl2 "do-translate-param-proc-appl1/10")
-			(let* ((tvar-bindings
-				(hfield-ref deductions 'bindings))
-			       (r-inst-type
-				(bind-type-vars binder tvar-bindings
-						inst-type1))
-			       (r-type-var-values
-				(get-type-var-values-from-deductions2
-				 new-type-vars deductions))
-			       (r-arg-types
-				(bind-type-vars binder tvar-bindings
-						new-arg-types)))
-			  (translate-normal-param-proc-appl2
-			   binder e-proc arglist r-inst-type
-			   r-arg-types
-			   r-type-var-values pure? always-returns?
-			   never-returns?
-			   l-default-params)))
-		       ((not-null? l-default-params)
-			(let* ((tvar-bindings
-				(map cons new-type-vars l-default-params))
-			       (r-inst-type
-				(bind-type-vars binder tvar-bindings
-						inst-type1))
-			       (r-arg-types
-				(bind-type-vars binder tvar-bindings
-						new-arg-types)))
-			  (translate-normal-param-proc-appl2
-			   binder e-proc arglist r-inst-type
-			   r-arg-types
-			   l-default-params pure? always-returns?
-			   never-returns?
-			   l-default-params)))
-		       (else
-			(let ((s-name
-			       (if (not-null? (hfield-ref
-					       e-proc
-					       'address))
-				   (hfield-ref
-				    (hfield-ref e-proc
-						'address)
-				    'source-name)
-				   '())))
-			  (raise (list 'did-not-deduce-all-type-vars
-				       s-name
-				       (cons 'actual-type new-arg-types)
-				       (cons 'declared-type arg-list-type0)
-				       (cons
-					'bindings
-					(hfield-ref deductions 'bindings))
-				       (cons 'needed new-type-vars))))))
-		      (cond
-		       ((not type-dispatched?)
-			(dwl2 "do-translate-param-proc-appl1/11")
-			(make-hrecord
-			 <proc-appl>
-			 tc-object
-			 #f #f '()
-			 pure? #f #t
-			 '() always-returns? never-returns?
-			 e-proc arglist '() static-arg-types
-			 #f
-			 l-default-params))
-		       ((all-tvars-correct? deductions new-type-vars)
-			(dwl2 "do-translate-param-proc-appl1/12")
-			(let* ((tvar-bindings
-				(hfield-ref deductions 'bindings))
-			       (r-inst-type
-				(bind-type-vars binder tvar-bindings
-						inst-type1))
-			       (r-type-var-values
-				(get-type-var-values-from-deductions2
-				 new-type-vars deductions))
-			       (r-arg-types
-				(bind-type-vars binder tvar-bindings
-						new-arg-types)))
-			  (translate-param-proc-appl-inside-param
-			   binder e-proc arglist r-inst-type
-			   r-arg-types
-			   r-type-var-values pure?
-			   always-returns? never-returns?
-			   l-default-params)))
-		       ((not-null? l-default-params)
-			(let* ((tvar-bindings
-				(map cons new-type-vars l-default-params))
-			       (r-inst-type
-				(bind-type-vars binder tvar-bindings
-						inst-type1))
-			       (r-arg-types
-				(bind-type-vars binder tvar-bindings
-						new-arg-types)))
-			  (translate-param-proc-appl-inside-param
-			   binder e-proc arglist r-inst-type
-			   r-arg-types
-			   l-default-params pure?
-			   always-returns? never-returns?
-			   l-default-params)))
-		       (else
-			(dwl2 "do-translate-param-proc-appl1/13")
-			(let* ((tvar-bindings
-				(hfield-ref deductions 'bindings))
-			       (r-inst-type
-				(bind-type-vars binder tvar-bindings
-						inst-type1))
-			       (result-type
-				(tno-field-ref r-inst-type
-					       'type-result)))
-			  (if (contains-only-specified-unbound-tvars-fwd?
-			       result-type
-			       fixed-tvars)
-			      (make-hrecord
-			       <proc-appl>
-			       result-type
-			       #t #f '()
-			       pure? #f #t
-			       '() always-returns? never-returns?
-			       e-proc arglist '() static-arg-types
-			       #f
-			       l-default-params)
-			      (make-hrecord
-			       <proc-appl>
-			       tc-object
-			       #f #f '()
-			       pure? #f #t
-			       '() always-returns? never-returns?
-			       e-proc arglist '() static-arg-types
-			       #f
-			       l-default-params)))))))))))))
+		      (raise 'illegal-nondispatched-type)
+		      (make-hrecord
+		       <proc-appl>
+		       tc-object
+		       #f #f '()
+		       pure? #f #t
+		       '() always-returns? never-returns?
+		       e-proc arglist '()
+		       (map get-expr-type arglist)
+		       #t
+		       l-default-params))
+		  (do-translate-param-proc-appl0 binder inside-param-def?
+						 type-dispatched?
+						 e-proc arglist
+						 l-default-params
+						 ppc inst-type
+						 pure? always-returns?
+						 never-returns?)))))))
 
 
 (define (do-translate-param-proc-appl binder inside-param-def?
@@ -1055,20 +990,6 @@
 
 (define (check-param-proc-type binder e-proc arglist l-default-params)
   (dwl2 "check-param-proc-type")
-
-  ;; TBR
-  (set! gl-counter7 (+ gl-counter7 1))
-  (dwl2 gl-counter7)
-  ;; (if (= gl-counter7 34)
-  ;;     (begin
-  ;; 	(dvar1-set! binder)
-  ;; 	(dvar2-set! e-proc)
-  ;; 	(dvar3-set! arglist)
-  ;; 	(raise 'stop-pp)))
-  (if (not-null? (hfield-ref e-proc 'address))
-      (dwl2 (hfield-ref (hfield-ref e-proc 'address) 'source-name)))
-
-
   (assert (hrecord-is-instance? binder <binder>))
   (assert (is-entity? e-proc))
   (assert (and (list? arglist)
@@ -1319,15 +1240,7 @@
 ;; as argument types
 (define (do-translate-normal-genproc-appl binder
 					  gen-proc arglist arg-types)
-  (dwl3 "do-translate-normal-genproc-appl")
-
-  ;; TBR
-  ;; (if gl-flag15?
-  ;;     (begin
-  ;; 	(dvar1-set! arglist)
-  ;; 	(dvar2-set! arg-types)
-  ;; 	(raise 'stop-*2)))
-
+  (dwl1 "do-translate-normal-genproc-appl")
   (let ((selection (select-best-method
 		    binder
 		    arg-types
@@ -1335,7 +1248,7 @@
     ;; MIETI puhtauden periytyminen metodeilla.
     (cond
      ((null? selection)
-      (dwl3 "do-translate-normal-genproc-appl/1")
+      (dwl1 "do-translate-normal-genproc-appl/1")
       (if (or-map? contains-free-tvars-fwd? arg-types)
 	  (make-general-gen-proc-appl binder gen-proc arglist
 				      tc-object #f #f #f #f #f)
@@ -1347,7 +1260,7 @@
       (let ((best-match (car selection)))
 	(assert (pair? best-match))
 	(let ((mtc (cdr best-match)))
-	  (dwl3 "do-translate-normal-genproc-appl/3")
+	  (dwl1 "do-translate-normal-genproc-appl/3")
 	  (dvar1-set! best-match)
 	  (if (not (is-tc-simple-proc? mtc))
 	      (raise 'invalid-procedure-1)
@@ -1356,7 +1269,7 @@
 		     (and (tno-field-ref mtc 'pure-proc?)
 			  (and-map? is-pure-entity? arglist)))
 		    (method (car best-match)))
-		(dwl3 "do-translate-normal-genproc-appl/4")
+		(dwl1 "do-translate-normal-genproc-appl/4")
 		(if (or (tno-field-ref mtc 'static-method?)
 			(is-exact-match? binder arg-types mtc))
 		    (make-gen-proc-appl-best-match
@@ -1374,7 +1287,7 @@
 		       #t
 		       pure? always-returns? never-returns? #t))))))))
      (else
-      (dwl3 "do-translate-normal-genproc-appl/5")
+      (dwl1 "do-translate-normal-genproc-appl/5")
       ;; Method ambiguity is not a compile time error (?).
       ;; Methods defined after calling the generic procedure
       ;; may remove the (run time) ambiguity.
@@ -1409,7 +1322,7 @@
 
 (define (do-translate-genproc-appl binder
 				   gen-proc arglist arg-types)
-  (dwl3 "do-translate-genproc-appl")
+  (dwl1 "do-translate-genproc-appl")
   (assert (is-binder? binder))
   (strong-assert (is-target-object? gen-proc))
   (assert (list? arglist))
@@ -1423,25 +1336,27 @@
       (let ((l-signatures (pick-signatures arg-types)))
 	(if (null? l-signatures)
 	    (do-translate-normal-genproc-appl binder gen-proc arglist arg-types)
-	    (if (not (contains-free-tvars-general? arglist))
-		(let* ((p-match (match-call-with-signatures binder
-							    gen-proc
-							    arg-types
-							    l-signatures))
-		       (tt-match (car p-match))
-		       (proc-encountered? (cdr p-match)))
-		  (cond
-		   ((not-null? tt-match)
-		    (translate-matched-proc-appl-with-sgn binder gen-proc
-							  arglist
-							  tt-match #f))
-		   (proc-encountered?
-		    (raise 'type-mismatch-with-signatures))
-		   (else
-		    (do-translate-normal-genproc-appl binder gen-proc
-						      arglist arg-types))))
-		(translate-unresolved-proc-appl-with-sgn binder gen-proc
-							 arglist #f))))))
+	    (begin
+	      (dwl1 "do-translate-genproc-appl/2")
+	      (if (not (contains-free-tvars-general? arglist))
+		  (let* ((p-match (match-call-with-signatures binder
+							      gen-proc
+							      arg-types
+							      l-signatures))
+			 (tt-match (car p-match))
+			 (proc-encountered? (cdr p-match)))
+		    (cond
+		     ((not-null? tt-match)
+		      (translate-matched-proc-appl-with-sgn binder gen-proc
+							    arglist
+							    tt-match #f))
+		     (proc-encountered?
+		      (raise 'type-mismatch-with-signatures))
+		     (else
+		      (do-translate-normal-genproc-appl binder gen-proc
+							arglist arg-types))))
+		  (translate-unresolved-proc-appl-with-sgn binder gen-proc
+							   arglist #f)))))))
 
 
 (define (get-general-generic-class with-result? appl-pure?)
@@ -1639,63 +1554,65 @@
    ((eqv? proc tp-class-of)
     (translate-class-of-appl binder proc arguments))
    ((eqv? proc tp-is-subtype)
-    (translate-is-subtype-appl binder proc arguments))
+    (translate-is-subtype-appl binder proc arguments
+			       type-check? inside-param-def?))
    ((eqv? proc tp-is-instance)
-    (translate-is-instance-appl binder proc arguments))
-   ;;	   ((eqv? proc tp-call/cc)
-   ;;	    (let ((binder (param-cache->binder param-cache)))
-   ;;	      (translate-call/cc binder proc arguments)))
+    (translate-is-instance-appl binder proc arguments
+				type-check? inside-param-def?))
    ((eqv? proc tp-tuple-ref)
-    (translate-tuple-ref-appl binder arguments type-check?))
+    (translate-tuple-ref-appl binder arguments
+			      type-check? inside-param-def?))
    ((eqv? proc tp-tuple-type-with-tail)
     (translate-tuple-type-with-tail-appl binder
-					 proc arguments type-check?))
+					 proc arguments inside-param-def?))
    ((eqv? proc tp-make-vector)
     (translate-make-vector-appl binder proc arguments
-				#f #f type-check?))
+				#f #f type-check? inside-param-def?))
    ((eqv? proc tp-make-mutable-vector)
     (translate-make-vector-appl binder proc arguments
-				#t #f type-check?))
+				#t #f type-check? inside-param-def?))
    ((eqv? proc tp-make-value-vector)
     (translate-make-vector-appl binder proc arguments
-				#f #t type-check?))
+				#f #t type-check? inside-param-def?))
    ((eqv? proc tp-make-mutable-value-vector)
     (translate-make-vector-appl binder proc arguments
-				#t #t type-check?))
+				#t #t type-check? inside-param-def?))
    ((eqv? proc tp-vector)
-    (translate-vector-appl binder proc arguments #f #f type-check?))
+    (translate-vector-appl binder proc arguments #f #f
+			   type-check? inside-param-def?))
    ((eqv? proc tp-value-vector)
-    (translate-vector-appl binder proc arguments #f #t type-check?))
+    (translate-vector-appl binder proc arguments #f #t
+			   type-check? inside-param-def?))
    ((eqv? proc tp-mutable-vector)
     (translate-vector-appl binder proc arguments #t #f
-			   type-check?))
+			   type-check? inside-param-def?))
    ((eqv? proc tp-mutable-value-vector)
     (translate-vector-appl binder proc arguments #t #t
-			   type-check?))
+			   type-check? inside-param-def?))
    ((eqv? proc tp-cast-vector)
     (translate-cast-vector-appl binder proc arguments #f #f
-				type-check?))
+				type-check? inside-param-def?))
    ((eqv? proc tp-cast-mutable-vector)
     (translate-cast-vector-appl binder proc arguments #t #f
-				type-check?))
+				type-check? inside-param-def?))
    ((eqv? proc tp-cast-value-vector)
     (translate-cast-vector-appl binder proc arguments #f #t
-				type-check?))
+				type-check? inside-param-def?))
    ((eqv? proc tp-cast-mutable-value-vector)
     (translate-cast-vector-appl binder proc arguments #t #t
-				type-check?))
+				type-check? inside-param-def?))
    ((eqv? proc tp-cast-vector-metaclass)
     (translate-cast-vector-metaclass binder proc arguments #f #f
-				     type-check?))
+				     type-check? inside-param-def?))
    ((eqv? proc tp-cast-mutable-vector-metaclass)
     (translate-cast-vector-metaclass binder proc arguments #t #f
-				     type-check?))
+				     type-check? inside-param-def?))
    ((eqv? proc tp-cast-value-vector-metaclass)
     (translate-cast-vector-metaclass binder proc arguments #f #t
-				     type-check?))
+				     type-check? inside-param-def?))
    ((eqv? proc tp-cast-mutable-value-vector-metaclass)
     (translate-cast-vector-metaclass binder proc arguments #t #t
-				     type-check?))
+				     type-check? inside-param-def?))
    ((eqv? proc tp-field-ref)
     (raise 'internal-invalid-field-ref))
    ((eqv? proc tp-field-set)
@@ -1766,11 +1683,6 @@
        (is-t-instance? binder
 		       (get-entity-type to-param-proc)
 		       tpc-param-proc)))
-
-  ;; TBR
-;;  (set! gl-counter10 (+ gl-counter10 1))
-;;  (dwl2 gl-counter10)
-
   (let* ((to-ppc (get-entity-type to-param-proc))
 	 (inst-type (tno-field-ref to-ppc 'type-contents)))
     (strong-assert (and (not-null? inst-type) (is-entity? inst-type)))
@@ -1796,14 +1708,6 @@
 	     (tmp2 (begin (dwl2 "translate-param-proc-dispatch/1-2") 0))
 	     (r-type-var-values (get-type-var-values-from-deductions2
 				 type-variables deductions)))
-
-	;; TBR
-	(dvar1-set! type-variables)
-	(dvar2-set! deductions)
-	(dvar3-set! inst-type)
-	(dvar4-set! bound-inst-type)
-;;	(raise 'stop-dispatch)
-
 	(dwl2 "translate-param-proc-dispatch/2")
 	(if type-check?
 	    (let* ((tt-actual-arglist
@@ -1826,25 +1730,6 @@
 		  (raise 'param-proc-dispatch-type-mismatch)))
 	    (begin
 	      (dwl2 "translate-param-proc-dispatch/4")
-	      ;; (let* ((l-free-tvars (get-all-free-tvars bound-inst-type))
-	      ;; 	     (type (if (null? l-free-tvars)
-	      ;; 		       bound-inst-type
-	      ;; 		       (make-param-proc-class-object
-	      ;; 			"unnamed"
-	      ;; 			l-free-tvars
-	      ;; 			bound-inst-type))))
-
-		;; TBR
-;;		(dvar1-set! type)
-;;		(raise 'stop1)
-
-
-	      ;; TBR
-	      (dvar1-set! r-type-var-values)
-	      (dvar2-set! bound-inst-type)
-	      (dvar3-set! deductions)
-;;	      (raise 'stop-dispatch2)
-
 	      (make-hrecord <expr-param-proc-dispatch>
 			    bound-inst-type
 			    #t
