@@ -64,24 +64,25 @@
   (dwl4 "parse-internal-variable")
   (assert (hrecord-is-instance? compiler <compiler>))
   (assert (hrecord-is-instance? address-env <address-environment>))
-  (strong-assert (and (list? expr) (= (length expr) 7)
+  (strong-assert (and (list? expr) (= (length expr) 8)
 		      (eq? (car expr) 'internal-variable)))
   (dwl4 "parse-internal-variable/1")
   (let ((p-address (list-ref expr 1))
 	(p-type (list-ref expr 2))
-	(exact-type? (list-ref expr 3))
-	(read-only? (list-ref expr 4))
-	(volatile? (list-ref expr 5))
-	(forward-decl? (list-ref expr 6)))
+	(type-dispatched? (list-ref expr 3))
+	(exact-type? (list-ref expr 4))
+	(read-only? (list-ref expr 5))
+	(volatile? (list-ref expr 6))
+	(forward-decl? (list-ref expr 7)))
     (dwl4 "parse-internal-variable/2")
     (let* ((module (get-current-module-name compiler))
 	   (r-address (parse-address module p-address))
 	   (r-type (parse-repr-fwd compiler address-env p-type))
 	   (var
-	    (make-normal-variable8 r-address r-type
-				   exact-type? read-only? volatile?
-				   forward-decl?
-				   '() #f)))
+	    (make-normal-variable10 r-address r-type type-dispatched?
+				    exact-type? read-only? volatile?
+				    forward-decl?
+				    '() #f)))
       (dwl4 "parse-internal-variable EXIT")
       var)))
 
