@@ -129,19 +129,6 @@
 	      binder param-type type-args visited))
 	 (repr-new (car pr)))
     (set! l-inst (append l-inst (cdr pr)))
-
-    ;; TBR
-    ;; (dw3 "apti hash (input): ")
-    ;; (dwl3 (hashq repr 1000000))
-    ;; (dw3 "apti hash (1): ")
-    ;; (dwl3 (hashq repr-new 1000000))
-
-    ;; (if (has-name? param-type ':my-tree)
-    ;; 	(begin
-    ;; 	  (dvar1-set! repr)
-    ;; 	  (dvar2-set! repr-new)
-    ;; 	  (raise 'stop-apti)))
-
     (dwli2 "do-bind-type-vars-apti/1")
     (dwli2 (debug-get-string repr-new))
     (if (eq? repr repr-new)
@@ -150,11 +137,8 @@
 	(let ((a (assq repr-new visited)))
 	  (if a
 	      (begin
-;;		(dwli2 "do-bind-type-vars-apti/2")
-;;		(dwli2 (debug-get-string (cdr a)))
 		(cons (cdr a) l-inst))
 	      (begin
-;;		(dwli2 "do-bind-type-vars-apti/3")
 		(cons repr-new l-inst)))))))
 
 
@@ -345,7 +329,8 @@
 		     (d2wli 'tvar-binding "do-bind-type-vars00/6-4")
 		     (if (and
 			  (hfield-ref repr 'runtime-arglist-typecheck?)
-			  (null? (hfield-ref repr 'static-arg-types)))
+			  (or (hfield-ref binder 'inside-param-proc?)
+			      (null? (hfield-ref repr 'static-arg-types))))
 			 (bind-type-vars-param-proc-runtime-fwd
 			  binder type-vars repr l-new-visited)
 			 (bind-type-vars-param-proc-fwd
@@ -475,8 +460,7 @@
   ;;     (if (hrecord? repr)
   ;; 	  (dwc (hrecord-type-name-of repr))
   ;; 	  (dwc "?")))
-  ;; (dwli-newline)
-
+  ;; (dwli-newline)  
   (let ((old-indent gl-indent)
 	(i-counter gl-counter8))
     (set! gl-indent (+ gl-indent 1))
@@ -620,20 +604,6 @@
   (check-bindings tvars)
   (assert (is-t-type-loop? repr))
   (assert (list? visited))
-
-  (dvar1-set! repr)
-
-  ;; TBR
-  ;; (set! gl-counter4 (+ gl-counter4 1))
-  ;; (dwl4 gl-counter4)
-
-  ;; TBR
-  ;; (dvar1-set! binder)
-  ;; (dvar2-set! bind)
-  ;; (dvar3-set! tvars)
-  ;; (dvar4-set! (list repr visited))
-  ;; (raise 'stop-loop)
-
   (let ((iter-var (tno-field-ref repr 'tvar))
 	(subtype-list0 (tno-field-ref repr 'x-subtypes))
 	(iter-expr (tno-field-ref repr 'x-iter-expr)))
@@ -659,11 +629,6 @@
 	(let* ((tt-param (get-uniform-list-param binder subtype-list))
 	       (l-tvars2 (cons (cons iter-var tt-param) tvars))
 	       (br3 (do-bind-type-vars2 binder l-tvars2 iter-expr visited)))
-
-	  ;; TBR
-	  ;; (dvar1-set! br3)
-	  ;; (raise 'stop-uniform1)
-
 	  (cons (make-tt-uniform-list (car br3))
 		(append (cdr br) (cdr br3)))))
        (else
@@ -683,16 +648,6 @@
 		 (raise 'internal-error-in-type-loop-type-list)))))
 	  ;;	    (dwli2 "bind-type-loop/3")
 	  (dwli2 "bind-type-loop/6")
-
-	  ;; TBR
-	  (dwi2 "iter-expr = ")
-	  (dwli2 (debug-get-string iter-expr))
-	  (dwi2 "subexprs = ")
-	  (dwli2 (debug-get-string subexprs))
-
-	  ;; TBR
-	  ;;	    (if gl-flag12? (set! gl-flag13? #t))
-
 	  (let* ((br2
 		  (do-bind-loop bind tvars iter-var
 				subexprs
@@ -1028,23 +983,6 @@
   (d2wli 'tvar-binding "do-bind-type-vars-param-proc")
   (set! gl-counter11 (+ gl-counter11 1))
   (d2wli 'tvar-binding gl-counter11)
-
-  ;; TBR
-  ;; (if (= gl-counter11 12)
-  ;;     (raise 'stop12))
-
-  ;;  (dwl1 "do-bind-type-vars-param-proc")
-  ;;  (dwl1 (hfield-ref binder 'optimize-copying?))
-
-  ;; TBR
-  ;; (if (has-name? param-proc 'nonpure-iter-for-each1)
-  ;;     (begin
-  ;; 	(dvar1-set! new-type-vars)
-  ;; 	(dvar2-set! new-tvar-values)
-  ;; 	(dvar3-set! inst-type)
-  ;; 	(dvar4-set! fixed-tvars)
-  ;; 	(raise 'stop-iter1)))
-
   (let ((instantiations '())
 	(i-counter gl-counter11))
     (if (and (is-pure-entity? param-proc)
@@ -1082,36 +1020,6 @@
 	    ;;   (cons repr-appl '()))
 	    (let* ((tmp1 (begin (d2wli 'tvar-binding "do-bind-type-vars-param-proc/5-2") 0))
 		   (param-cache (get-param-cache binder))
-
-		   ;; ;; TBR
-		   ;; (tmp2 (begin (d2wli 'tvar-binding "do-bind-type-vars-param-proc/5-3")
-		   ;; 		    (set! gl-counter11 (+ gl-counter11 1))
-		   ;; 		    (d2wli 'tvar-binding gl-counter11)
-		   ;; 		    0))
-
-		   ;; ;; TBR
-		   ;; (tmp2-1 (begin (d2wli 'tvar-binding "do-bind-type-vars-param-proc/5-3-1")
-		   ;; 		    (if (= gl-counter11 8)
-		   ;; 			(begin
-		   ;; 			  (dvar1-set! param-cache)
-		   ;; 			  (dvar2-set! param-proc)
-		   ;; 			  (dvar3-set! new-tvar-values)
-		   ;; 			  (raise 'stop-rec2)))
-		   ;; 		    0))
-
-
-		   ;; TBR
-		   ;; (tmp2-2 (begin
-		   ;; 		 (dw3 "param proc inst (1): ")
-		   ;; 		 (dwl3 (get-param-proc-instance-name
-		   ;; 			param-proc new-tvar-values))
-		   ;; 		 (set! gl-counter17 (+ gl-counter17 1))
-		   ;; 		 (dwl3 gl-counter17)
-		   ;; 		 0))
-		   ;; (if (= gl-counter17 15)
-		   ;; 	(raise 'stop17))
-
-
 		   ;; It is essential that new-tvar-values is in the same order
 		   ;; as new-type-vars.
 		   (old-binding (param-cache-fetch2
@@ -1133,44 +1041,6 @@
 				     (hfield-ref param-proc 'address)
 				     'source-name)
 				    '()))
-			       
-			       ;; TBR
-			       ;; (tmp4
-			       ;;  (begin
-			       ;; 	(if flag1?
-			       ;; 	    (begin
-			       ;; 	      (dvar1-set! (list
-			       ;; 			   binder
-			       ;; 			   address
-			       ;; 			   bound-param-proc
-			       ;; 			   new-tvar-values
-			       ;; 			   cycles))
-			       ;; 	      (raise 'stop-inst-2)))
-			       ;; 	0))
-			       ;; (tmp5
-			       ;;  (begin
-			       ;;    (if (= (hfield-ref address 'number) 207)
-			       ;; 	  (begin
-			       ;; 	    (set! gl-flag13? #t)))
-			       ;;    0))
-
-
-			       ;; TBR
-			       ;; (tmp5
-			       ;;  (begin
-			       ;;    (if (= (hfield-ref address 'number) 56)
-			       ;; 	  (begin
-			       ;; 	    (dwl2 "56 HEP")
-			       ;; 	    (dwl2 counter)
-			       ;; 	    (dvar1-set!
-			       ;; 	     (list
-			       ;; 	      binder address inst-type new-type-vars
-			       ;; 	      new-tvar-values cycles))
-			       ;; 	    (dvar2-set! fixed-tvars)
-			       ;;  	    (set! gl-show-indented-debug-info? #t)
-			       ;; 	    (raise 'stop56)))
-			       ;;    0))
-			       
 			       (proc-instance
 				(if #f
 				    (inst-make-param-proc-instance-w-params-fwd
@@ -1187,22 +1057,6 @@
 						    new-tvar-values obj)
 			  (set! instantiations
 				(append instantiations subinst))
-
-			  ;; TBR
-			  ;; 		      (if (= (hfield-ref address 'number)
-			  ;; 		             56)
-			  ;; 		          (begin
-			  ;; 			    (dvar1-set!
-			  ;; 			     (list binder obj param-proc new-tvar-values
-			  ;; 				   name cycles))
-			  ;; 			    (dvar2-set! (list new-arg-types inst-type))
-			  ;; 			    (dvar3-set! bound-args)
-			  ;; 			    (dwl2 param?)
-			  ;; 			    (set! gl-flag10? #t)
-			  ;; ;;			    (set! gl-show-indented-debug-info? #t)
-			  ;; 			    (dwl2 counter)
-			  ;; 		            (raise 'stop56)))
-
 			  (if (and
 			       (is-known-object? param-proc)
 			       (not (hfield-ref param-proc 'incomplete?)))
@@ -1213,14 +1067,6 @@
 					 new-tvar-values name cycles))
 				       (inst-expr (car contents))
 				       (subinst2 (cadr contents)))
-
-				  ;; TBR
-				  ;; (if flag1?
-				  ;;     (begin
-				  ;;       (dvar1-set! obj)
-				  ;;       (dvar2-set! inst-expr)
-				  ;;       (raise 'stop-set)))
-
 				  (if (hfield-ref binder 'optimize-copying?)
 				      (if #f
 					  (set! instantiations
@@ -1238,41 +1084,10 @@
 				   binder 'decl-proc-instances
 				   (cons (list obj param-proc new-tvar-values)
 					 (hfield-ref binder 'decl-proc-instances)))))
-			  
-			  ;; TBR
-			  ;; (if flag1?
-			  ;;     (begin
-			  ;;       (dvar1-set! obj)
-			  ;;       (raise 'stop-obj)))
-			  
-			  ;; (if (= (hfield-ref address 'number) 296)
-			  ;; 	  (begin
-			  ;; 	    (raise 'stop296)))
-
 			  obj)))
 		   (new-repr
 		    (begin
 		      (d2wli 'tvar-binding "do-bind-type-vars-param-proc/6")
-
-		      ;; TBR
-		      (d2wli 'tvar-binding (hfield-ref (hfield-ref new-instance 'address) 'number))
-
-		      ;; TBR
-		      ;; (if (= (hfield-ref (hfield-ref new-instance 'address) 'number)
-		      ;; 	   215)
-		      ;; 	(begin
-		      ;; 	  (dvar1-set! new-instance)
-		      ;; 	  (raise 'stop215)))
-
-
-		      ;; TBR
-		      ;; (if (has-name? param-proc 'nonpure-iter-for-each1)
-		      ;; 	  (begin
-		      ;; 	    (dvar1-set! binder)
-		      ;; 	    (dvar2-set! new-instance)
-		      ;; 	    (dvar3-set! bound-args)
-		      ;; 	    (raise 'stop-iter2)))
-
 		      (if (hfield-ref binder 'optimize-copying?)
 			  (translate-simple-proc-appl-expression1-fwd
 			   binder
@@ -1287,18 +1102,6 @@
 			   (get-entity-type new-instance)
 			   new-tvar-values)))))
 	      (debug-write-line-and-indent "do-bind-type-vars-param-proc/8")
-
-	      ;; TBR
-	      ;; (d2wli 'tvar-binding "proc hash:")
-	      ;; (d2wli 'tvar-binding (hashq (get-entity-type (hfield-ref new-repr 'proc))
-	      ;; 		1000000))
-
-	      ;; TBR
-	      ;; (if (= i-counter 13)
-	      ;; 	  (begin
-	      ;; 	    (dvar1-set! new-repr)
-	      ;; 	    (raise 'stop13)))
-
 	      (cons new-repr instantiations))))))
 
 
@@ -1309,35 +1112,13 @@
   (check-bindings type-vars)
   (assert (hrecord-is-instance? repr <proc-appl>))
   (assert (list? cycles))
-
-  (set! gl-counter9 (+ gl-counter9 1))
-  (d2wli 'tvar-binding gl-counter9)
-
-  ;; TBR
-  (let ((address (hfield-ref (hfield-ref repr 'proc) 'address)))
-    (if (not-null? address)
-	(begin
-	  (d2wli 'tvar-binding (hfield-ref address 'module))
-	  (d2wli 'tvar-binding (hfield-ref address 'source-name))
-	  (d2wli 'tvar-binding (hfield-ref address 'number)))
-	(d2wli 'tvar-binding "null")))
-  ;; (if (= gl-counter9 36)
-  ;;     (begin
-  ;; 	(dvar1-set! repr)
-  ;; 	(raise 'stop36)))
-  ;; (if (has-name? (hfield-ref repr 'proc) 'car)
-  ;;     (begin
-  ;; 	(dvar1-set! repr)
-  ;; 	(raise 'stop-appl2)))
-
   (let ((old-indent gl-indent)
 	(i-counter gl-counter9))
     (set! gl-indent (+ gl-indent 1))
 
-    ;; TBR
-    (dwi2 "indent: ")
-    (dwc2 gl-indent)
-    (dwli-newline2)
+    ;; (dwi2 "indent: ")
+    ;; (dwc2 gl-indent)
+    ;; (dwli-newline2)
 
     (let* ((instantiations '())
 	   (args (hfield-ref repr 'arglist))
@@ -1711,18 +1492,6 @@
   (check-bindings type-vars)
   (assert (hrecord-is-instance? repr <proc-appl>))
   (assert (list? cycles))
-
-  ;; TBR
-  (dwl2 "bind-param-prim-proc-appl ENTER")
-;;  (set! gl-counter24 (+ gl-counter24 1))
-;;  (dwl2 gl-counter24)
-;;  (dvar1-set! type-vars)
-;;  (raise 'stop24)
-  ;; (if (has-name? (hfield-ref repr 'proc) 'car)
-  ;;     (begin
-  ;; 	(dvar1-set! repr)
-  ;; 	(raise 'stop-appl)))
-
   (let ((old-indent gl-indent))
     (set! gl-indent (+ gl-indent 1))
     (let* ((proc (hfield-ref repr 'proc))
@@ -1743,15 +1512,6 @@
 	   (l-new-default-params (map* do-inst l-default-params))
 	   (new-subreprs (list proc new-arglist l-new-default-params))
 	   (new-repr (cwb-param-proc-appl binder repr new-subreprs #t)))
-
-      ;; TBR
-      ;; (if (and gl-flag16?
-      ;; 	       (has-name? proc 'append))
-      ;; 	  (begin
-      ;; 	    (dvar1-set! new-repr)
-      ;; 	    (dvar2-set! repr)
-      ;; 	    (raise 'stop16)))
-
       (set! gl-indent old-indent)
       (d2wli 'tvar-binding "bind-param-prim-proc-appl EXIT")
       (cons new-repr instantiations))))
@@ -1772,14 +1532,6 @@
 	       (instance (create-fwd-ref-for-instance-fwd
 			  bound-param-class address))
 	       (l-old-overwrite (hfield-ref binder 'l-overwrite)))
-	  
-	  ;; TBR
-	  ;; (if (and (not-null? address)
-	  ;; 	   (= (hfield-ref address 'number) 304))
-	  ;;     (begin
-	  ;; 	(dvar1-set! binder)
-	  ;; 	(raise 'stop304-1)))
-
 	  (hfield-set! binder 'l-overwrite
 		       (cons instance l-old-overwrite))
 	  (param-cache-add-binding! param-cache bound-param-class params
@@ -1803,16 +1555,6 @@
   (assert (is-binder? binder))
   (assert (list? type-vars))
 
-  (set! gl-counter4 (+ gl-counter4 1))
-  (dwli2 gl-counter4)
-  ;; (if (= gl-counter4 78)
-  ;;     (begin
-  ;; 	(dvar1-set! binder)
-  ;; 	(dvar2-set! type-vars)
-  ;; 	(dvar3-set! repr)
-  ;; 	(dvar4-set! cycles)
-  ;; 	(raise 'stop-pc)))
-
   (check-bindings type-vars)
 
   (assert (is-t-param-class-instance? repr))
@@ -1824,20 +1566,11 @@
 	     'could-not-handle-param-class-args
 	     repr type-vars)))
 
-  ;; TBR
-;;  (dvar1-set! repr)
-;;  (raise 'stop1)
-
   (let ((instantiations '())
 	(param-class (get-entity-type repr))
 	(args (tno-field-ref repr 'l-tvar-values))
 	(old-indent gl-indent))
     (set! gl-indent (+ gl-indent 1))
-
-    ;; TBR
-    (dwli2 (debug-get-string param-class))
-    (dwli2 (debug-get-string args))
-
     (dwli2 "bind-type-vars-param-class/2")
     (let ((subresult-pc (do-bind-type-vars2 binder type-vars param-class
 					    cycles)))
@@ -1956,20 +1689,9 @@
   (assert (is-t-field? field))
   (let ((old-indent gl-indent))
     (set! gl-indent (+ gl-indent 1))
-    (dwli2 (tno-field-ref field 's-name))
-    (dwli2 "original type:")
-    (dwli2 (debug-get-string (tno-field-ref field 'type)))
-
-    ;; TBR
-    ;; (if (and gl-flag13?
-    ;; 	     (eq? (tno-field-ref field 'sym-name) 'contents))
-    ;; 	(begin
-    ;; 	  (dvar1-set! binder)
-    ;; 	  (dvar2-set! type-vars)
-    ;; 	  (dvar3-set! field)
-    ;; 	  (dvar4-set! cycles)
-    ;; 	  (raise 'stop13-2)))
-
+    ;; (dwli2 (tno-field-ref field 's-name))
+    ;; (dwli2 "original type:")
+    ;; (dwli2 (debug-get-string (tno-field-ref field 'type)))
     (let* ((inst '())
 	   (to-type (tno-field-ref field 'type))
 	   (has-init-value? (tno-field-ref field 'has-init-value?))
@@ -2015,8 +1737,8 @@
 		      (make-field sym-name to-new-type acc-read acc-write
 				  has-init-value? obj-new-init-value)
 		      inst)))))
-	  (dwli2 "result type:")
-	  (dwli2 (debug-get-string (tno-field-ref (car result) 'type)))
+	  ;; (dwli2 "result type:")
+	  ;; (dwli2 (debug-get-string (tno-field-ref (car result) 'type)))
 	  (set! gl-indent old-indent)
 	  (dwli2 "bind-type-vars-for-field EXIT")
 	  result))))))
@@ -2151,21 +1873,8 @@
   (assert (list? type-var-values))
   (assert (and-map? is-target-object? type-var-values))
   (let ((instantiations '()))
-
     ;; Should we have strong-assert?
     (assert (= (length type-var-values) (length tvars)))
-
-    ;; TBR
-    (if gl-flag13?
-    	(begin
-    	  (dvar1-set! binder)
-    	  (dvar2-set! (map cons tvars type-var-values))
-    	  (dvar3-set! inst-type)
-    	  (dvar4-set! cycles)
-;;    	  (set! gl-show-indented-debug-info? #t)
-	  (dwli2 "HEP 13")))
-;;    	  (raise 'stop13)))
-
     (let* ((type-var-bindings (map cons tvars type-var-values))
 	   (bind-result1 (do-bind-type-vars2
 			  binder type-var-bindings
@@ -2226,14 +1935,8 @@
   (assert (and-map? is-target-object? type-var-values))
   (assert (or (symbol? name) (null? name)))
 
-  (d2wli 'tvar-binding "inst-set-param-proc-instance-contents")
-  (d2wli 'tvar-binding name)
-
-  ;; TBR
-  ;; (if (eq? name 'gen-assoc)
-  ;;     (begin
-  ;; 	(dvar1-set! param-proc)
-  ;;  	(raise 'stop-appl)))
+  ;; (d2wli 'tvar-binding "inst-set-param-proc-instance-contents")
+  ;; (d2wli 'tvar-binding name)
 
   (let* ((instantiations '())
 	 (to-ppc (get-entity-type param-proc))
@@ -2242,39 +1945,8 @@
 	 (expr-old-cur-proc (hfield-ref binder 'expr-cur-proc)))
     ;; Ordinary assert might be enough.
     (strong-assert (= (length type-var-values) (length tvars)))
-    (dvar1-set! param-proc)
-    (dvar2-set! name)
-
-    (dwl2 "*1*")
     (set-cur-param-proc! binder param-proc)
-    (dvar1-set! param-proc)
     (hfield-set! binder 'expr-cur-proc (tno-field-ref param-proc 'x-value-expr))
-    (dwl2 "*2*")
-
-    ;; TBR
-    ;; (if (= gl-counter11 36)
-    ;; 	(begin
-    ;; 	  (dvar1-set! param-proc)
-    ;; 	  (raise 'stop36-2)))
-    ;; TBR
-    ;; (if (= (hfield-ref (hfield-ref param-proc 'address) 'number) 67)
-    ;;  	(begin
-    ;;   	  (dvar1-set! param-proc)
-    ;;   	  (raise 'stop67)))
-    ;; (dw3 "param proc inst (2): ")
-    ;; (dwl3 (get-param-proc-instance-name
-    ;; 	   param-proc type-var-values))
-    ;; (set! gl-counter18 (+ gl-counter18 1))
-    ;; (dwl3 gl-counter18)
-
-    ;; (if gl-flag15?
-    ;; 	(begin
-    ;; 	  (dvar1-set! tvars)
-    ;; 	  (dvar2-set! type-var-values)
-    ;; 	  (dvar3-set! param-proc)
-    ;; 	  (dvar4-set! binder)
-    ;; 	  (raise 'stop15)))
-
     (let* ((type-var-bindings (map cons tvars type-var-values))
 	   ;; (inst-type (tno-field-ref to-ppc 'inst-type))
 	   (value-expr (tno-field-ref param-proc 'x-value-expr))
@@ -2310,33 +1982,6 @@
       inst-set-param-proc-instance-contents)
 
 
-;; (define (inst-bind-param-proc-class binder type-vars repr cycles)
-;;   ;; Note: type-vars contains bindings instead of plain type variables.
-;;   (assert (is-binder? binder))
-;;   (assert (list? type-vars))
-;;   (assert (and (list? cycles) (and-map? is-object-marker? cycles)))
-;;   (let* ((inst-type (tno-field-ref repr 'inst-type))
-;; 	 (bound-tvars (tno-field-ref repr 'type-vars))
-;; 	 (new-bindings
-;; 	  (filter (lambda (binding)
-;; 		    (not (member (car binding) bound-tvars
-;; 				 type-variable=?)))
-;; 		  type-vars))
-;; 	 (bind-result
-;; 	  (do-bind-type-vars2 binder new-bindings inst-type cycles))
-;; 	 (new-inst-type (car bind-result))
-;; 	 (instantiations (cdr bind-result))
-;; 	 (new-ppc
-;; 	  (if (eqv? new-inst-type inst-type)
-;; 	      repr
-;; 	      (clone-with-branches-fwd 
-;; 	       binder
-;; 	       repr
-;; 	       (list new-inst-type)
-;; 	       #t))))
-;;     (cons new-ppc instantiations)))
-
-
 (define (inst-bind-param-proc-class binder type-vars repr cycles)
   (assert (is-binder? binder))
   ;; Note: type-vars contains bindings instead of plain type variables.
@@ -2395,14 +2040,14 @@
 
 (define (inst-check-let-variable binder to-decl-type ent-init-expr s-name)
   (assert (is-binder? binder))
-  (assert (is-target-object? to-decl-type))
+  (assert (or (null? to-decl-type) (is-target-object? to-decl-type)))
   (assert (is-entity? ent-init-expr))
   (assert (symbol? s-name))
   (let ((to-expr-type (get-entity-type ent-init-expr)))
     (cond
-     ;; Should we test only one of the types?
-     ((or (contains-type-variables-fwd? to-decl-type)
-	  (contains-type-variables-fwd? to-expr-type))
+     ((null? to-decl-type) #t)
+     ((or (contains-free-tvars-fwd? to-decl-type)
+	  (contains-free-tvars-fwd? to-expr-type))
       #t)
      ((and (hfield-ref binder 'type-check?)
 	   (not (is-t-subtype? binder to-expr-type to-decl-type)))
@@ -2422,8 +2067,11 @@
       var-old
       (make-normal-variable6
        (hfield-ref var-old 'address)
-       to-new-type
-       #t
+       (if (not-null? to-new-type)
+	   to-new-type
+	   (get-entity-type expr-new-init))
+       (or (not-null? to-new-type)
+	   (entity-type-dispatched? expr-new-init))
        #f
        read-only?
        (hfield-ref var-old 'volatile?)
@@ -2439,7 +2087,8 @@
    (is-target-object? (car letvar))
    (is-normal-variable? (cadr letvar))
    (is-entity? (caddr letvar))
-   (is-target-object? (list-ref letvar 3))
+   (or (null? (list-ref letvar 3))
+       (is-target-object? (list-ref letvar 3)))
    (is-entity? (list-ref letvar 4))
    (is-t-primitive-object? (list-ref letvar 5))))
 
@@ -2514,16 +2163,6 @@
   (assert (list? type-vars))
   (check-bindings type-vars)
   (assert (hrecord-is-instance? repr <let-expression>))
-
-  ;; (set! gl-counter22 (+ gl-counter22 1))
-  ;; (dwl2 gl-counter22)
-
-  ;; TBR
-  ;; (if (= gl-counter22 10)
-  ;;     (begin
-  ;; 	(dvar1-set! repr)
-  ;; 	(raise 'stop10)))
-
   (let* ((i-counter gl-counter22)
 	 (recursive? (hfield-ref repr 'recursive?))
 	 (order? (hfield-ref repr 'order?))
@@ -2544,35 +2183,15 @@
 		   (set! instantiations
 			 (append instantiations (cdr bind-result)))
 		   (car bind-result))))
-
-	 ;; TBR
-	 ;; (tmp1 (begin
-	 ;; 	 (dwl2 "inst-bind-let-expression/2-1")
-	 ;; 	 (if (= i-counter 4)
-	 ;; 	     (begin
-	 ;; 	       (dvar1-set! bind)
-	 ;; 	       (dvar2-set! init-exprs)
-	 ;; 	       (raise 'stop4)))
-	 ;; 	 0))
-
 	 (l-new-init-exprs (map* bind init-exprs))
-	 (tmp2 (begin (dwl2 "inst-bind-let-expression/2-2") 0))
 	 (l-new-declared-types (map* bind declared-types)))
     (dwl2 "inst-bind-let-expression/2")
+    (dwl2 l-s-names)
     (for-each (lambda (to-decl-type ent-init-expr s-name)
 		(inst-check-let-variable binder to-decl-type ent-init-expr
 					 s-name))
 	      l-new-declared-types l-new-init-exprs l-s-names)
     (dwl2 "inst-bind-let-expression/3")
-
-    ;; TBR
-    ;; (if (= i-counter 2)
-    ;; 	(begin
-    ;; 	  (dvar1-set! l-old-vars)
-    ;; 	  (dvar2-set! (list declared-types init-exprs))
-    ;; 	  (dvar3-set! (list l-new-declared-types l-new-init-exprs))
-    ;; 	  (raise 'stop22-2)))
-
     (let* ((l-new-vars
 	    (map* (lambda
 		      (var-old to-old-type expr-old-init to-new-type
@@ -2583,6 +2202,7 @@
 		  l-old-vars declared-types init-exprs
 		  l-new-declared-types l-new-init-exprs))
 	   (l-new-bindings (map (lambda (bind-object? var init-expr)
+				  (dwl2 bind-object?)
 				  (if bind-object? init-expr var))
 				l-bind-object l-new-vars l-new-init-exprs))
 	   (letvars1
@@ -2592,8 +2212,6 @@
 	    (let ((result
 		   (cond
 		    ((and (not recursive?) (not order?))
-		     (dwl2 "inst-bind-let-expression/4")
-		     (dwl2 i-counter)
 		     (cons letvars1 '()))
 		    ((and (not recursive?) order?)
 		     (raise 'let*-not-builtin))
@@ -2621,8 +2239,9 @@
 	    (raise (list 'let:initializer-none
 			 (cons 's-name s-name-none)))))
       (dwl2 "inst-bind-let-expression/6")
+      (set! gl-counter27 (+ gl-counter27 1))
+      (dwl2 gl-counter27)
       (let* ((old-body (hfield-ref repr 'body))
-	     (tmp1 (begin (dwl2 "inst-bind-let-expression/7") 0))
 	     (l-subst (inst-make-let-subst old-bindings l-new-bindings))
 	     (new-body
 	      (let* ((bindings (append l-subst type-vars))
@@ -2630,8 +2249,6 @@
 		      (do-bind-type-vars2 binder bindings old-body cycles)))
 		(set! instantiations (append instantiations (cdr result)))
 		(car result)))
-	     (tmp2 (begin (dwl2 "inst-bind-let-expression/8") 0))
-
 	     (new-body-pure? (is-pure-entity? new-body))
 	     (final-init-exprs-pure?
 	      (and-map? is-pure-entity? l-new-init-exprs))
@@ -2659,7 +2276,6 @@
 			    new-body)))
 	(dwl2 i-counter)
 	(dwl2 "inst-bind-let-expression EXIT")
-	;; (raise 'stop-let)
 	(cons result instantiations)))))
 
 
@@ -2692,36 +2308,14 @@
 
 
 (define (inst-bind-procedure binder type-vars repr cycles)
-;;  (dwl2 "inst-bind-procedure")
-;;  (dwl2 (hfield-ref binder 'type-check?))
   (assert (is-binder? binder))
   (assert (list? type-vars))
   (check-bindings type-vars)
   (assert (hrecord-is-instance? repr <procedure-expression>))  
-  (d2wli 'tvar-binding "inst-bind-procedure/1")
-  (d2wli 'tvar-binding (hfield-ref binder 'type-check?))
-
-  (set! gl-counter26 (+ gl-counter26 1))
-  (d2wli 'tvar-binding gl-counter26)
-
-  ;; TBR
-  ;; (let ((names (hfield-ref repr 'arg-names)))
-  ;;   (if (and (not-null? names)
-  ;; 	     (eq? (car names) 'lst-op-specs))
-  ;; 	(begin
-  ;; 	  (dvar1-set! repr)
-  ;; 	  (raise 'lst-stop-4))))
-
+  ;; (d2wli 'tvar-binding "inst-bind-procedure/1")
+  ;; (d2wli 'tvar-binding (hfield-ref binder 'type-check?))
   (let* ((instantiations '())
 	 (i-counter gl-counter26)
-;;	 (local15? gl-flag15?)
-
-	 ;; TBR
-	 ;; (local-flag?
-	 ;;  (if gl-flag10?
-	 ;;      (begin (set! gl-flag10? #f) #t)
-	 ;;      #f))
-
 	 (old-arg-descs (hfield-ref repr 'arg-descs))
 	 (new-arg-descs
 	  (map (lambda (arg-desc)
@@ -2733,10 +2327,6 @@
 		   (car bind-result)))
 	       old-arg-descs))
 	 (old-result-type (hfield-ref repr 'result-type))
-
-	 ;; TBR
-	 (tmp1 (begin (d2wli 'tvar-binding "inst-bind-procedure/1-1") 0))
-
 	 (new-result-type
 	  (let ((bind-result
 		 (do-bind-type-vars2 binder type-vars old-result-type
@@ -2747,48 +2337,17 @@
 	 (impl-arg-types
 	  (get-impl-arg-types binder new-arg-descs))
 	 (names (hfield-ref repr 'arg-names))
-
-	 ;,TBR
-	 ;; (tmp1-0 (begin
-	 ;; 	   (if (and (not-null? names)
-	 ;; 		    (eq? (car names) 'proc-2283))
-	 ;; 	       (begin
-	 ;; 		 (d2wli 'tvar-binding "arg HEP")
-	 ;; 		 (d2wli 'tvar-binding (hfield-ref gl-linker 'next-free-loc))))
-	 ;; 	   0))
-
 	 (alloc-var (hfield-ref binder 'allocate-variable))
 	 (new-arg-vars (make-new-argument-variables
 			binder alloc-var names impl-arg-types))
 	 (old-arg-vars (hfield-ref repr 'arg-variables))
 	 (old-body (hfield-ref repr 'body))
-
-	 ;; TBR
-	 (tmp2 (begin (d2wli 'tvar-binding "inst-bind-procedure/1-2") 0))
-	 ;; (tmp2-1 (begin
-	 ;; 	   (if local15?
-	 ;; 	       (begin
-	 ;; 		 (dvar1-set! new-arg-vars)
-	 ;; 		 (dvar2-set! old-arg-vars)
-	 ;; 		 (dvar3-set! binder)
-	 ;; 		 (dvar4-set! old-body)
-	 ;; 		 (raise 'stop15-2)))
-	 ;; 	   0))
-
 	 (body1
 	  (if (and-map? eqv? new-arg-vars old-arg-vars)
 	      old-body
 	      (rebind-local-variables1-fwd
 	       binder old-body
 	       (map cons old-arg-vars new-arg-vars))))
-
-	 ;; TBR
-	 (tmp3 (begin (d2wli 'tvar-binding "inst-bind-procedure/1-3")
-		      (d2wli 'tvar-binding i-counter)
-		      0))
-	 ;; 	      (if gl-flag16? (set! gl-flag17? #t))
-	 ;; 	      0))
-
 	 (new-body
 	  ;; Field fixing? is always #f.
 	  (if (hfield-ref binder 'fixing?)
@@ -2797,31 +2356,6 @@
 		(set! instantiations (append instantiations (cdr result)))
 		(car result)))))
     (d2wli 'tvar-binding "inst-bind-procedure/2")
-
-    ;; TBR
-    ;; (set! gl-counter22 (+ gl-counter22 1))
-    ;; (dwl2 gl-counter22)
-    ;; (dvar1-set! repr)
-    ;; (raise 'stop22)
-
-    ;; TBR
-    ;; (if gl-flag16?
-    ;; 	(begin
-    ;; 	  (dvar1-set! new-body)
-    ;; 	  (dvar2-set! body1)
-    ;; 	  (dvar3-set! binder)
-    ;; 	  (dwl2 type-check?)
-    ;; 	  (dwl2 "HEP16")
-    ;; 	  (raise 'stop16)))
-
-    ;; TBR
-    ;; (if (and (not-null? names)
-    ;;  	     (eq? (car names) 'lst-op-specs))
-    ;;  	(begin
-    ;;  	  (dvar1-set! new-arg-descs)
-    ;;  	  (dvar2-set! old-arg-descs)
-    ;;  	  (raise 'lst-stop)))
-
     (if (or (not (= (length old-arg-descs) (length new-arg-descs)))
 	    (not (and-map? eqv? old-arg-descs new-arg-descs))
 	    (not (eqv? old-result-type new-result-type))
@@ -2911,15 +2445,6 @@
 			      (hfield-ref repr 's-name)))
 			 (l-module
 			  (hfield-ref repr 'l-module))
-			 ;; TBR
-			 ;; (tmp1 (begin
-			 ;; 	 (if (equal? names '(stack element))
-			 ;; 	     (begin
-			 ;; 	       (dvar1-set! new-arg-descs)
-			 ;; 	       (dvar2-set! new-arg-vars)
-			 ;; 	       (raise 'stop-push)))
-			 ;; 	 0))
-
 			 (result-expr
 			  (make-hrecord <procedure-expression>
 					proc-type
@@ -2943,28 +2468,8 @@
 					appl-always-returns?
 					appl-never-returns?
 					static-method?)))
-		    
-		    ;; TBR
-		    ;; (if local-flag?
-		    ;; 	(begin
-		    ;; 	  (dvar1-set! binder)
-		    ;; 	  (dvar2-set! type-vars)
-		    ;; 	  (dvar3-set! body1)
-		    ;; 	  (dvar4-set! cycles)
-		    ;; 	  (raise 'stop-proc1)))
-
 		    (dwl2 "inst-bind-procedure EXIT 1")
 		    (cons result-expr instantiations)))))))
 	(begin
-	  
-	  ;; TBR
-	  ;; (if local-flag?
-	  ;;     (begin
-	  ;; 	(dvar1-set! binder)
-	  ;; 	(dvar2-set! type-vars)
-	  ;; 	(dvar3-set! body1)
-	  ;; 	(dvar4-set! cycles)
-	  ;; 	(raise 'stop-proc2)))
-
 	  (dwl2 "inst-bind-procedure EXIT 2")
 	  (cons repr instantiations)))))
