@@ -65,6 +65,7 @@
   (list 'internal-variable
 	(theme-compile-address (hfield-ref var 'address))
 	(theme-compile-repr compiler env (get-entity-type var))
+	(hfield-ref var 'type-dispatched?)
 	(hfield-ref var 'exact-type?)
 	(hfield-ref var 'read-only?)
 	(hfield-ref var 'volatile?)
@@ -84,6 +85,7 @@
     (list 'internal-variable
 	  p-address
 	  p-type
+	  (hfield-ref var 'type-dispatched?)
 	  (hfield-ref var 'exact-type?)
 	  (hfield-ref var 'read-only?)
 	  (hfield-ref var 'volatile?)
@@ -97,6 +99,7 @@
   (list 'internal-variable-with-expr
 	(theme-compile-address (hfield-ref var 'address))
 	(theme-compile-repr compiler env (get-entity-type var))
+	(hfield-ref var 'type-dispatched?)
 	(hfield-ref var 'exact-type?)
 	(hfield-ref var 'read-only?)
 	(hfield-ref var 'volatile?)
@@ -778,7 +781,10 @@
 	(let* ((s-name (get-contents name))
 	       (p-address (theme-compile-address
 			   (hfield-ref variable 'address)))
-	       (t-type-decl (theme-compile-repr compiler env type-decl))
+	       (t-type-decl
+		(if (not-null? type-decl)
+		    (theme-compile-repr compiler env type-decl)
+		    '()))
 	       (t-bind-object? (is-t-true? bind-object))
 	       (t-init-expr (theme-compile-repr compiler env init-expr))
 	       (t-variable
