@@ -2244,6 +2244,11 @@
 (set! is-type-fwd? is-type?)
 
 
+(define (is-type2? binder obj)
+  (and (is-target-object? obj)
+       (is-t-subtype-fwd? binder (get-entity-type obj) tt-type)))
+
+			  
 (define (is-null-class-entity? ent)
   (and (is-entity? ent) (eq? (get-entity-value ent) tc-nil)))
 
