@@ -1228,11 +1228,15 @@
 	      t-arglist2)
 	(let* ((static-arg-types (hfield-ref repr 'static-arg-types))
 	       (t-static
-		(if (list? static-arg-types)
-		    (let ((t (map* comp static-arg-types)))
-		      `(list ,@t))
-		    (comp static-arg-types))))
-	  (list '_i_call-proc t-proc t-arglist2 `(cadr ,t-static))))))
+		(cond
+		 ((null? static-arg-types) '(quote ()))
+		 ((list? static-arg-types)
+		  (if (>= (length static-arg-types) 2)
+		      (comp (cadr static-arg-types))
+		      '(quote ())))
+		 (else
+		  `(cadr ,(comp static-arg-types))))))
+	  (list '_i_call-proc t-proc t-arglist2 t-static)))))
 
 
 (define (tcomp-proc-appl linker repr)
@@ -1346,7 +1350,8 @@
     (assert (hrecord-is-instance? init-expr <entity>))
     (if (hfield-ref linker 'tcomp-inside-param-proc?)
 	(let ((type-check
-	       (if (entity-type-dispatched? init-expr)
+	       (if (or (null? type)
+		       (entity-type-dispatched? init-expr))
 		   (theme-target-compile linker init-expr)
 		   (generate-type-check
 		    linker
