@@ -406,10 +406,10 @@
 ;; For field-ref and field-set! only the procedures
 ;; needed by both the compiler and the linker
 ;; are defined here. The procedures needed by
-;; only the compiler are in theme-fields.scm.
+;; only the compiler are in theme-d-fields.scm.
 
 
-(define (translate-const-field-ref binder obj-expr field-name type-check?)
+(define (translate-const-field-ref binder obj-expr field-name)
   (dwl4 "translate-const-field-ref")
   (assert (is-binder? binder))
   (assert (is-entity? obj-expr))
@@ -449,7 +449,7 @@
 
 
 (define (translate-const-field-set binder obj-expr field-name
-				   value type-check?)
+				   value type-check? inside-param-def?)
   (dwl4 "translate-const-field-set")
   (assert (is-binder? binder))
   (assert (is-entity? obj-expr))
@@ -460,6 +460,7 @@
 		      (not (entity-type-is-none? value))))
   (dwl4 "translate-const-field-set/1")
   (let* ((type-expr (get-entity-type obj-expr))
+	 (type-dispatched? (entity-type-dispatched? obj-expr))
 	 (always-returns?
 	  (and (entity-always-returns? obj-expr)
 	       (entity-always-returns? value)))
@@ -467,7 +468,9 @@
 	  (or (entity-never-returns? obj-expr)
 	      (entity-never-returns? value))))
     (cond
-     ((contains-free-tvars-general-fwd? type-expr)
+     ((and type-check? (not inside-param-def?) (not type-dispatched?))
+      (raise 'illegal-undispatched-type))
+     ((or (not type-dispatched?) (contains-free-tvars-fwd? type-expr))
       (make-hrecord <field-set-expr>
 		    tt-none
 		    #t
@@ -503,7 +506,8 @@
 
 				  #f
 				  #f
-				  (not type-check?)
+				  (or (not type-check?)
+				      inside-param-def?)
 				  '()
 
 				  always-returns?
@@ -527,7 +531,7 @@
 
 
 (define (translate-tuple-type-with-tail-appl binder
-					     proc arguments type-check?)
+					     proc arguments inside-param-def?)
   (assert (is-binder? binder))
   (assert (is-entity? proc))
   (assert (and (list? arguments)
@@ -537,7 +541,7 @@
       ;; What about purity?
       (let ((tuple-part (car arguments))
 	    (tail-part (cadr arguments)))
-	(if (or (not type-check?)
+	(if (or inside-param-def?
 		(contains-free-tvars-general-fwd? tuple-part)
 		(contains-free-tvars-general-fwd? tail-part)
 		(and (is-t-instance? binder tuple-part tt-type)
@@ -557,8 +561,9 @@
       (let* ((r-value (car arguments))
 	     (r-type (get-entity-type r-value)))
 	(if (and (not-null? r-type)
-		 (and-map? is-known-pure-entity? arguments)
-		 (not (contains-free-tvars-general-fwd? r-type))
+		 (is-known-pure-entity? r-value)
+		 (entity-type-dispatched? r-value)
+		 ;; (not (contains-free-tvars-general-fwd? r-type))
 		 (is-t-instance? binder r-type tc-class)
 		 (not (tno-field-ref r-type 'inheritable?)))
 	    r-type
@@ -569,13 +574,14 @@
 		  (static-arg-types (list (get-entity-type r-value))))
 	      (make-hrecord <proc-appl>
 			    tc-class
+			    ;; Should we set type-dispatched? = #t?
 			    type-dispatched?
 			    ;; Should we set exact-type? = #t?
 			    #f
 			    '()
 			    pure-args?
 			    #f
-			    #t
+			    (not type-dispatched?)
 			    '()
 			    always-returns?
 			    never-returns?
@@ -588,51 +594,65 @@
       (raise 'class-of:invalid-number-of-arguments)))
 
 
-(define (translate-is-subtype-appl binder proc arguments)
+(define (translate-is-subtype-appl binder proc arguments type-check?
+				   inside-param-def?)
   (assert (is-binder? binder))
   (assert (is-entity? proc))
   (assert (and (list? arguments)
 	       (and-map? is-entity? arguments)))
+  (assert (boolean? inside-param-def?))
   (check-no-none-arguments arguments 'is-subtype?)
+  (dwl1 "translate-is-subtype-appl")
   (if (= (length arguments) 2)
-      (let ((r-type1 (car arguments))
-	    (r-type2 (cadr arguments)))
-	(if (and  
-	     (and-map? is-known-pure-entity? arguments)
-	     (not (contains-free-tvars-general-fwd? r-type1))
-	     (not (contains-free-tvars-general-fwd? r-type2)))
-	    (make-primitive-object
-	     tc-boolean
-	     (is-t-subtype? binder
-			    (get-entity-value r-type1)
-			    (get-entity-value r-type2)))
-	    (let ((pure-args? (and-map? is-pure-entity? arguments))
-		  (static-arg-types (map get-entity-type arguments)))
-	      (make-hrecord <proc-appl>
-			    tc-boolean
-			    #t
-			    #t
-			    '()
-			    pure-args?
-			    #f
-			    #t
-			    '()
-			    ;; No need to check proc here for
-			    ;; special procedures.
-			    (and (entity-always-returns? r-type1)
-				 (entity-always-returns? r-type2))
-			    (or (entity-never-returns? r-type1)
-				(entity-never-returns? r-type2))
-			    tp-is-subtype
-			    arguments
-			    '()
-			    static-arg-types
-			    #f
-			    '()))))
+      (let* ((r-type1 (car arguments))
+	     (r-type2 (cadr arguments))
+	     (free-tvars?
+	      (or
+	       (contains-free-tvars-general-fwd? r-type1)
+	       (contains-free-tvars-general-fwd? r-type2))))
+	(cond
+	 ((and type-check? (not inside-param-def?) free-tvars?)
+	  (raise 'illegal-free-type-variables))
+	 ((and (not free-tvars?) 
+	       (and-map? is-known-pure-entity? arguments))
+	  (if (and (is-type2? binder r-type1)
+		   (is-type2? binder r-type2))
+	      (make-primitive-object
+	       tc-boolean
+	       (is-t-subtype? binder r-type1 r-type2))
+	      (raise 'is-subtype?:not-a-type)))
+	 (else
+	  (let ((pure-args? (and-map? is-pure-entity? arguments))
+		(static-arg-types (map get-entity-type arguments))
+		(correct-types? 
+		 (and (is-type-repr2? binder r-type1)
+		      (is-type-repr2? binder r-type2))))
+	    (make-hrecord <proc-appl>
+			  tc-boolean
+			  #t
+			  #t
+			  '()
+			  pure-args?
+			  #f
+			  (not correct-types?)
+			  '()
+			  ;; No need to check proc here for
+			  ;; special procedures.
+			  (and (entity-always-returns? r-type1)
+			       (entity-always-returns? r-type2))
+			  (or (entity-never-returns? r-type1)
+			      (entity-never-returns? r-type2))
+			  tp-is-subtype
+			  arguments
+			  '()
+			  static-arg-types
+			  (not correct-types?)
+			  '())))))
       (raise 'is-subtype?:invalid-number-of-arguments)))
 
 
-(define (translate-is-instance-appl binder proc arguments)
+(define (translate-is-instance-appl binder proc arguments
+				    type-check? inside-param-def?)
   (assert (is-binder? binder))
   (assert (is-entity? proc))
   (assert (and (list? arguments)
@@ -642,21 +662,34 @@
       (let* ((r-object (car arguments))
 	     (r-type (cadr arguments))
 	     (r-object-type (get-entity-type r-object))
-	     (contains-free-tvars?
-	      (or (contains-free-tvars-general-fwd? r-object-type)
-		  (contains-free-tvars-general-fwd? r-type))))
+	     (type-dispatched? (entity-type-dispatched? r-object))
+	     ;; (contains-free-tvars?
+	     ;;  (or (contains-free-tvars-general-fwd? r-object-type)
+	     ;; 	  (contains-free-tvars-general-fwd? r-type))))
+	     (free-tvars?
+	      (contains-free-tvars-fwd? r-type)))
 	(cond
+	 ((and type-check? (not inside-param-def?) (not type-dispatched?))
+	  (raise 'is-instance:invalid-object-type))
+	 ((and type-check? (not inside-param-def?) free-tvars?)
+	  (raise 'invalid-free-type-variables))
 	 ((and
 	   (and-map? is-known-pure-entity? arguments)
-	   (not contains-free-tvars?))
-	  (make-primitive-object
-	   tc-boolean
-	   (is-t-subtype? binder
-			  (get-entity-value r-object-type)
-			  (get-entity-value r-type))))
+	   type-dispatched?
+	   (not free-tvars?))
+	  ;; The following check should never fail.
+	  (assert (is-type2? binder r-object-type))
+	  (if (is-type2? binder r-type)
+	      (make-primitive-object
+	       tc-boolean
+	       (is-t-subtype? binder
+			      r-object-type
+			      r-type))
+	      (raise 'is-instance?:not-a-type)))
 	 (else
 	  (let ((pure-args? (and-map? is-pure-entity? arguments))
-		(static-arg-types (map get-expr-type arguments)))
+		(static-arg-types (map get-expr-type arguments))
+		(correct-type? (is-type-repr2? binder r-type)))
 	    (make-hrecord <proc-appl>
 			  tc-boolean
 			  #t
@@ -664,7 +697,7 @@
 			  '()
 			  #t
 			  #f
-			  #t
+			  (not correct-type?)
 			  '()
 			  (and (entity-always-returns? r-object)
 			       (entity-always-returns? r-type))
@@ -674,12 +707,13 @@
 			  arguments
 			  '()
 			  static-arg-types
-			  #f
+			  (not correct-type?)
 			  '())))))
       (raise 'is-instance?:invalid-number-of-arguments)))
 
 
-(define (translate-tuple-ref-appl binder arguments type-check?)
+(define (translate-tuple-ref-appl binder arguments
+				  type-check? inside-param-def?)
   (dwl4 "translate-tuple-ref-appl")
   (assert (is-binder? binder))
   (strong-assert (and (list? arguments)
@@ -693,22 +727,23 @@
 	 (pure-args? (and-map? is-pure-entity? arguments))
 	 (static-arg-types (map get-entity-type arguments))
 	 (type-dispatched? (hfield-ref r-tuple 'type-dispatched?)))
-    ;;	 (contains-free-tvars?
-    ;;	  (contains-free-tvars-general-fwd? r-type)))
-
-    ;; TBR
-    (dwl4 (hfield-ref r-index 'obj-prim-contents))
-    (dwl4 type-check?)
-    (dwl4 contains-free-tvars?)
-
+;;	 (contains-free-tvars? (contains-free-tvars-fwd? r-type)))
     (cond
-     ((and type-check? (not tuple-type?))
-      (raise 'tuple-ref:not-a-tuple))
-     ((not (and (is-t-subtype? binder (get-entity-type r-index) tc-integer)
-		(is-t-primitive-object? r-index)))
+     ((and type-check? (not inside-param-def?)
+	   ;;	   (not contains-free-tvars?)
+	   type-dispatched?
+	   (not tuple-type?))
+      (raise (list 'tuple-ref:not-a-tuple (cons 'tt r-type))))
+     ((and type-check? (not inside-param-def?)
+	   (not (and (is-t-subtype? binder (get-entity-type r-index)
+				    tc-integer)
+		     (is-t-primitive-object? r-index))))
       (raise 'tuple-ref:index-not-an-integer-constant))
      ;;     ((or (not type-check?) contains-free-tvars?)
-     ((not type-dispatched?)
+     ((or (not type-dispatched?)
+	  (and inside-param-def?
+	       (or (contains-free-tvars-fwd? r-type)
+		   (contains-free-tvars-fwd? (get-entity-type r-index)))))
       (dwl4 "translate-tuple-ref-appl/1")
       (make-hrecord <proc-appl>
 		    tc-object
@@ -726,8 +761,8 @@
 		    tp-tuple-ref
 		    arguments
 		    '()
-		    static-arg-types
-		    #f
+		    '()
+		    #t
 		    '()))
      (else
       (dwl4 "translate-tuple-ref-appl/2")
@@ -758,7 +793,7 @@
 
 (define (translate-make-vector-appl binder
 				    proc arguments mutable? eq-by-value?
-				    type-check?)
+				    type-check? inside-param-def?)
   (dwl3 "translate-make-vector-appl")
   (assert (is-binder? binder))
   (assert (is-entity? proc))
@@ -767,6 +802,7 @@
   (assert (boolean? mutable?))
   (assert (boolean? eq-by-value?))
   (assert (boolean? type-check?))
+  (assert (boolean? inside-param-def?))
   (check-no-none-arguments arguments
 			   (if mutable?
 			       (if eq-by-value?
@@ -783,26 +819,25 @@
 	     (r-size-type (get-entity-type r-size))
 	     (free-tvars-type? (contains-free-tvars-general-fwd? r-type))
 	     (free-tvars-size? (contains-free-tvars-general-fwd? r-size-type))
-	     (free-tvars-element? (contains-free-tvars-general-fwd?
-				   r-value-type)))
+	     (element-type-dispatched?
+	      (entity-type-dispatched? r-element-value)))
 	(cond
-	 ;; ((entity-is-none1? binder r-type)
-	 ;;  (raise 'make-x-vector:none-element-type))
+	 ((and type-check? (entity-is-none1? binder r-type))
+	  (raise 'make-x-vector:none-element-type))
 	 ;; ((entity-is-none1? binder r-value-type)
 	 ;;  (raise 'make-x-vector:none-value-type))
-	 ((and type-check?
-	       (not free-tvars-type?)
+	 ((and type-check? (not inside-param-def?)
 	       (not (is-t-instance? binder r-type tt-type)))
 	  (raise 'make-x-vector:invalid-type))
-	 ((and type-check?
-	       (not free-tvars-size?)
+	 ((and type-check? (not inside-param-def?)
 	       (not (is-t-subtype? binder r-size-type tc-integer)))
 	  (dvar1-set! r-size-type)
 	  (dvar2-set! arguments)
 	  (raise 'make-x-vector:invalid-size))
-	 ((and type-check?
-	       (not free-tvars-type?)
-	       (not free-tvars-element?)
+	 ((and type-check? (not inside-param-def?)
+	       (not element-type-dispatched?))
+	  (raise 'illegal-undispatched-type))
+	 ((and type-check? (not inside-param-def?)
 	       (not (is-t-subtype? binder r-value-type r-type)))
 	  (raise 'make-x-vector:element-type-mismatch))
 	 (else
@@ -815,8 +850,16 @@
 			 (translate-value-vector-expression0 r-type))
 		     (if mutable?
 			 (translate-mutable-vector-expression0 r-type)
-			 (translate-vector-expression0 r-type)))))
-	    (dwl3 "translate-make-vector-appl/1")
+			 (translate-vector-expression0 r-type))))
+		(correct-types?
+		 (or (not inside-param-def?)
+		     (and
+		      element-type-dispatched?
+		      (not free-tvars-type?)
+		      (not free-tvars-size?)
+		      (is-t-instance? binder r-type tt-type)
+		      (is-t-subtype? binder r-size-type tc-integer)
+		      (is-t-subtype? binder r-value-type r-type)))))
 	    (make-hrecord <proc-appl>
 			  result-type
 			  #t
@@ -824,7 +867,7 @@
 			  '()
 			  pure-args?
 			  #f
-			  (not type-check?)
+			  (not correct-types?)
 			  '()
 			  (and-map? entity-always-returns? arguments)
 			  (or-map? entity-never-returns? arguments)
@@ -838,13 +881,14 @@
 			  arguments
 			  '()
 			  static-arg-types
-			  #f
+			  (not correct-types?)
 			  '())))))
       (raise 'make-x-vector:invalid-number-of-arguments)))
 
 
 (define (translate-vector-appl binder
-			       proc arguments mutable? eq-by-value? type-check?)
+			       proc arguments mutable? eq-by-value?
+			       type-check? inside-param-def?)
   (dwl4 "translate-vector-appl")
   (assert (is-binder? binder))
   (assert (is-entity? proc))
@@ -853,6 +897,7 @@
   (assert (boolean? mutable?))
   (assert (boolean? eq-by-value?))
   (assert (boolean? type-check?))
+  (assert (boolean? inside-param-def?))
   (check-no-none-arguments arguments
 			   (if mutable?
 			       (if eq-by-value?
@@ -863,17 +908,26 @@
 				   'vector)))
   (if (null? arguments)
       (raise 'x-vector:invalid-number-of-arguments)
-      (let ((element-types (map get-entity-type (cdr arguments)))
-	    (result-element-type (car arguments)))
+      (let* ((element-types (map get-entity-type (cdr arguments)))
+	     (type-dispatched?
+	      (and-map? entity-type-dispatched? (cdr arguments)))
+	     (result-element-type (car arguments))
+	     (free-tvars-type?
+	      (contains-free-tvars-general-fwd? result-element-type)))
 	(cond
-	 ;; ((entity-is-none1? binder result-element-type)
-	 ;;  (raise 'x-vector:none-element-type))
+	 ((and type-check? (entity-is-none1? binder result-element-type))
+	  (raise 'x-vector:none-element-type))
 	 ;; ((or-map? (lambda (type)
 	 ;; 	     (entity-is-none1? binder type))
 	 ;; 	   element-types)
 	 ;;  (raise 'x-vector:none-value))
-	 ((and type-check?
-	       (not (contains-free-tvars-general-fwd? result-element-type))
+	 ((and type-check? (not inside-param-def?)
+	       (not type-dispatched?))
+	  (raise 'illegal-undispatched-type))
+	 ((and type-check? (not inside-param-def?)
+	       (not (is-t-instance? binder result-element-type tt-type)))
+	  (raise 'x-vector:invalid-type))
+	 ((and type-check? (not inside-param-def?)
 	       (not (and-map? (lambda (type)
 				(is-t-subtype?
 				 binder
@@ -898,7 +952,19 @@
 			 (translate-vector-expression0
 			  result-element-type))))
 		(static-arg-types (map get-entity-type arguments))
-		(pure-args? (and-map? is-pure-entity? arguments)))
+		(pure-args? (and-map? is-pure-entity? arguments))
+		(correct-types?
+		 (or (not inside-param-def?)
+		     (and
+		      type-dispatched?
+		      (not free-tvars-type?)
+		      (is-t-instance? binder result-element-type tt-type)
+		      (and-map? (lambda (type)
+				  (is-t-subtype?
+				   binder
+				   type
+				   result-element-type))
+				element-types)))))
 	    (make-hrecord <proc-appl>
 			  result-type
 			  #t
@@ -906,7 +972,7 @@
 			  '()
 			  pure-args?
 			  #f
-			  (not type-check?)
+			  (not correct-types?)
 			  '()
 			  (and-map? entity-always-returns? arguments)
 			  (or-map? entity-never-returns? arguments)
@@ -920,7 +986,7 @@
 			  arguments
 			  '()
 			  static-arg-types
-			  #f
+			  (not correct-types?)
 			  '())))))))
 
 
@@ -998,12 +1064,13 @@
 
 (define (translate-cast-vector-appl binder
 				    proc arguments mutable? value-vector?
-				    type-check?)
+				    type-check? inside-param-def?)
   (assert (is-binder? binder))
   (assert (is-entity? proc))
   (assert (and (list? arguments)
 	       (and-map? is-entity? arguments)))
   (assert (boolean? type-check?))
+  (assert (boolean? inside-param-def?))
   (check-no-none-arguments arguments
 			   (if mutable?
 			       (if value-vector?
@@ -1015,79 +1082,39 @@
   (if (= (length arguments) 2)
       (let* ((source-vector (cadr arguments))
 	     (source-vector-type (get-entity-type source-vector))
-	     (vec? (is-vector-metaclass? binder source-vector-type)))
-	(if (and type-check? (not vec?))
-	    (raise 'cast-x-vector:invalid-vector)
-	    (let* ((target-element-type (car arguments))
-		   (target-type (get-cast-target-type mutable? value-vector?
-						      target-element-type))
-		   (pure-args? (and-map? is-pure-entity? arguments))
-		   (static-arg-types (map get-entity-type arguments))
-		   (opt? (and vec?
-			      (is-t-subtype? binder
-					     (get-vector-class-element-type
-					      source-vector-type)
-					     target-element-type)))
-		   (target-procedure (get-cast-obj mutable? value-vector?
-						   opt?)))
-	      (make-hrecord <proc-appl>
-			    target-type
-			    #t
-			    ;; Think about exact-type?
-			    ;; The runtime type of the target shall be
-			    ;; (:vector target-element-type)
-			    ;; even though the runtime types of the vector
-			    ;; elements may be subtypes of target-element-type.
-			    #t
-			    '()
-			    pure-args?
-			    #f
-			    (not type-check?)
-			    '()
-			    (and-map? entity-always-returns? arguments)
-			    (or-map? entity-never-returns? arguments)
-			    target-procedure
-			    arguments
-			    '()
-			    static-arg-types
-			    #f
-			    '()))))
-      (begin
-	(dvar1-set! proc)
-	(dvar2-set! arguments)
-	(raise 'cast-x-vector:invalid-number-of-arguments))))
-
-
-(define (translate-cast-vector-metaclass binder
-					 proc arguments mutable? value-vector?
-					 type-check?)
-  (dwl4 "translate-cast-vector-metaclass")
-  (assert (is-binder? binder))
-  (assert (is-entity? proc))
-  (assert (and (list? arguments)
-	       (and-map? is-entity? arguments)))
-  (assert (boolean? type-check?))
-  (check-no-none-arguments arguments
-			   (if mutable?
-			       (if value-vector?
-				   'cast-mutable-value-vector-metaclass
-				   'cast-mutable-vector-metaclass)
-			       (if value-vector?
-				   'cast-value-vector-metaclass
-				   'cast-vector-metaclass)))
-  (dvar1-set! arguments)
-  (if (= (length arguments) 1)
-      (let* ((source-vector (car arguments))
-	     (source-vector-type (get-entity-type source-vector))
-	     (element-type (get-vector-class-element-type source-vector-type))
-	     (target-type (get-cast-target-type mutable? value-vector?
-						element-type))
-	     (pure-args? (and-map? is-pure-entity? arguments))
-	     (target-procedure (get-metacast-obj mutable? value-vector?))
-	     (static-arg-types (map get-entity-type arguments)))
-	(if (or
-	     (not type-check?)
-	     (is-vector-metaclass? binder source-vector-type))
+	     (vec? (is-vector-metaclass? binder source-vector-type))
+	     (target-element-type (car arguments))
+	     (type-dispatched?
+	      (and (entity-type-dispatched? source-vector)
+		   (entity-type-dispatched? target-element-type))))
+	(cond
+	 ((and type-check? (not inside-param-def?) (not type-dispatched?))
+	  (raise 'illegal-undispatched-type))
+	 ((and type-check? (not inside-param-def?) (not vec?))
+	  (raise 'cast-x-vector:not-a-vector))
+	 ((and type-check? (not inside-param-def?)
+	       (not (is-t-instance? binder target-element-type tt-type)))
+	  (raise 'cast-x-vector:invalid-target-type))
+	 (else
+	  (let* ((target-type (get-cast-target-type mutable? value-vector?
+						    target-element-type))
+		 (pure-args? (and-map? is-pure-entity? arguments))
+		 (static-arg-types (map get-entity-type arguments))
+		 (opt? (and vec?
+			    (is-t-subtype? binder
+					   (get-vector-class-element-type
+					    source-vector-type)
+					   target-element-type)))
+		 (target-procedure (get-cast-obj mutable? value-vector?
+						 opt?))
+		 (correct-types?
+		  (or (not inside-param-def?)
+		      (and type-dispatched?
+			   vec?
+			   (not (contains-free-tvars-general-fwd?
+				 target-element-type))
+			   (is-t-subtype? binder target-element-type
+					  tt-type)))))
 	    (make-hrecord <proc-appl>
 			  target-type
 			  #t
@@ -1100,7 +1127,7 @@
 			  '()
 			  pure-args?
 			  #f
-			  (not type-check?)
+			  (not correct-types?)
 			  '()
 			  (and-map? entity-always-returns? arguments)
 			  (or-map? entity-never-returns? arguments)
@@ -1108,10 +1135,81 @@
 			  arguments
 			  '()
 			  static-arg-types
+			  (not correct-types?)
+			  '())))))
+      (begin
+	(dvar1-set! proc)
+	(dvar2-set! arguments)
+	(raise 'cast-x-vector:invalid-number-of-arguments))))
+
+
+(define (translate-cast-vector-metaclass binder
+					 proc arguments mutable? value-vector?
+					 type-check? inside-param-def?)
+  (dwl4 "translate-cast-vector-metaclass")
+  (assert (is-binder? binder))
+  (assert (is-entity? proc))
+  (assert (and (list? arguments)
+	       (and-map? is-entity? arguments)))
+  (assert (boolean? type-check?))
+  (assert (boolean? inside-param-def?))
+  (check-no-none-arguments arguments
+			   (if mutable?
+			       (if value-vector?
+				   'cast-mutable-value-vector-metaclass
+				   'cast-mutable-vector-metaclass)
+			       (if value-vector?
+				   'cast-value-vector-metaclass
+				   'cast-vector-metaclass)))
+  (dvar1-set! arguments)
+  (if (= (length arguments) 1)
+      (let* ((source-vector (car arguments))
+	     (source-vector-type (get-entity-type source-vector))
+	     (type-dispatched? (entity-type-dispatched? source-vector))
+	     (vec? (is-vector-metaclass? binder source-vector-type))
+	     (target-procedure (get-metacast-obj mutable? value-vector?))
+	     (static-arg-types (map get-entity-type arguments))
+	     (pure-args? (is-pure-entity? source-vector)))
+	(cond
+	 ((and type-check? (not inside-param-def?) (not vec?))
+	  (raise 'cast-vector-metaclass:not-a-vector))
+	 ((and type-check? (not inside-param-def?) (not type-dispatched?))
+	  (raise 'illegal-undispatched-type))
+	 (else
+	  (let ((target-type
+		 (if vec?
+		     (let ((element-type (get-vector-class-element-type
+					  source-vector-type)))
+		       (get-cast-target-type mutable? value-vector?
+					     element-type))
+		     tc-object))
+		(correct-types?
+		 (or (not inside-param-def?)
+		     (and type-dispatched?
+			  vec?
+			  (not (contains-free-tvars-fwd?
+				source-vector-type))))))
+	    (make-hrecord <proc-appl>
+			  target-type
+			  type-dispatched?
+			  ;; Think about exact-type?
+			  ;; The runtime type of the target shall be
+			  ;; (:vector target-element-type)
+			  ;; even though the runtime types of the vector
+			  ;; elements may be subtypes of target-element-type.
+			  #t
+			  '()
+			  pure-args?
 			  #f
-			  '())
-	    (begin
-	      (dvar1-set! source-vector)
-	      (raise 'cast-x-vector-metaclass:invalid-vector))))
+			  (not correct-types?)
+			  '()
+			  (and-map? entity-always-returns? arguments)
+			  (or-map? entity-never-returns? arguments)
+			  target-procedure
+			  arguments
+			  '()
+			  static-arg-types
+			  (not correct-types?)
+			  '())))))
       (raise 'cast-x-vector-metaclass:invalid-number-of-arguments)))
 
